import os
from typing import Any
from unittest.mock import Mock

from coinmetrics.api_client import CoinMetricsClient, requests
from coinmetrics._catalogs import (
    CatalogAssetsData,
    CatalogAssetAlertsData,
    CatalogAssetPairsData,
    CatalogExchangeAssetsData,
    CatalogExchangesData,
    CatalogIndexesData,
    CatalogInstitutionsData,
    CatalogMarketsData,
    CatalogMetricsData,
    CatalogAssetPairCandlesData,
    CatalogMarketTradesData,
)
from coinmetrics._data_collection import DataCollection
from coinmetrics._typing import (
    DataRetrievalFuncType,
    UrlParamTypes,
    Dict,
    DataReturnType,
    DataFrameType,
)

try:
    import pandas as pd  # type: ignore

    DataFrameType = pd.core.frame.DataFrame
except ImportError:
    pd = None
    DataFrameType = Any

catalog_assets_test_data = [
    {
        "asset": "btc",
        "full_name": "Bitcoin",
        "metrics": [
            {
                "metric": "TestMetric",
                "frequencies": [
                    {
                        "frequency": "1b",
                        "min_time": "2009-01-03T18:15:05.000000000Z",
                        "max_time": "2020-06-08T20:22:17.000000000Z",
                        "min_height": "0",
                        "max_height": "10",
                        "min_hash": "0abc",
                        "max_hash": "0xyz",
                    },
                    {
                        "frequency": "1d",
                        "min_time": "2009-01-03T00:00:00.000000000Z",
                        "max_time": "2020-06-07T00:00:00.000000000Z",
                        "community": True,
                    },
                ],
            }
        ],
        "exchanges": ["binance", "coinbase", "kraken"],
        "markets": [
            "binance-btc-usdt-spot",
            "binance-eth-btc-spot",
            "coinbase-btc-usd-spot",
            "coinbase-eth-btc-spot",
            "kraken-btc-usd-spot",
        ],
    }
]

catalog_candles_test_data = [
    {
        "frequencies": [
            {
                "frequency": "1m",
                "max_time": "2022-07-12T17:50:00.000000000Z",
                "min_time": "2020-12-26T01:00:00.000000000Z",
            },
            {
                "frequency": "5m",
                "max_time": "2022-07-12T17:45:00.000000000Z",
                "min_time": "2020-12-26T01:00:00.000000000Z",
            },
            {
                "frequency": "10m",
                "max_time": "2022-07-12T17:40:00.000000000Z",
                "min_time": "2020-12-26T01:00:00.000000000Z",
            },
            {
                "frequency": "15m",
                "max_time": "2022-07-12T17:30:00.000000000Z",
                "min_time": "2020-12-26T01:00:00.000000000Z",
            },
            {
                "frequency": "30m",
                "max_time": "2022-07-12T17:00:00.000000000Z",
                "min_time": "2020-12-26T01:00:00.000000000Z",
            },
            {
                "frequency": "1h",
                "max_time": "2022-07-12T16:00:00.000000000Z",
                "min_time": "2020-12-26T01:00:00.000000000Z",
            },
            {
                "frequency": "4h",
                "max_time": "2022-07-12T12:00:00.000000000Z",
                "min_time": "2020-12-26T00:00:00.000000000Z",
            },
            {
                "frequency": "1d",
                "max_time": "2022-07-11T00:00:00.000000000Z",
                "min_time": "2020-12-26T00:00:00.000000000Z",
            },
        ],
        "pair": "1inch-usd",
    },
    {
        "frequencies": [
            {
                "frequency": "1m",
                "max_time": "2022-07-12T17:50:00.000000000Z",
                "min_time": "2020-10-10T08:13:00.000000000Z",
            },
            {
                "frequency": "5m",
                "max_time": "2022-07-12T17:45:00.000000000Z",
                "min_time": "2020-10-10T08:10:00.000000000Z",
            },
            {
                "frequency": "10m",
                "max_time": "2022-07-12T17:40:00.000000000Z",
                "min_time": "2020-10-10T08:10:00.000000000Z",
            },
            {
                "frequency": "15m",
                "max_time": "2022-07-12T17:30:00.000000000Z",
                "min_time": "2020-10-10T08:00:00.000000000Z",
            },
            {
                "frequency": "30m",
                "max_time": "2022-07-12T17:00:00.000000000Z",
                "min_time": "2020-10-10T08:00:00.000000000Z",
            },
            {
                "frequency": "1h",
                "max_time": "2022-07-12T16:00:00.000000000Z",
                "min_time": "2020-10-10T08:00:00.000000000Z",
            },
            {
                "frequency": "4h",
                "max_time": "2022-07-12T12:00:00.000000000Z",
                "min_time": "2020-10-10T08:00:00.000000000Z",
            },
            {
                "frequency": "1d",
                "max_time": "2022-07-11T00:00:00.000000000Z",
                "min_time": "2020-10-10T00:00:00.000000000Z",
            },
        ],
        "pair": "aave-usd",
    },
    {
        "frequencies": [
            {
                "frequency": "1m",
                "max_time": "2022-07-12T17:50:00.000000000Z",
                "min_time": "2018-02-26T00:00:00.000000000Z",
            },
            {
                "frequency": "5m",
                "max_time": "2022-07-12T17:45:00.000000000Z",
                "min_time": "2018-02-26T00:00:00.000000000Z",
            },
            {
                "frequency": "10m",
                "max_time": "2022-07-12T17:40:00.000000000Z",
                "min_time": "2018-02-26T00:00:00.000000000Z",
            },
            {
                "frequency": "15m",
                "max_time": "2022-07-12T17:30:00.000000000Z",
                "min_time": "2018-02-26T00:00:00.000000000Z",
            },
            {
                "frequency": "30m",
                "max_time": "2022-07-12T17:00:00.000000000Z",
                "min_time": "2018-02-26T00:00:00.000000000Z",
            },
            {
                "frequency": "1h",
                "max_time": "2022-07-12T16:00:00.000000000Z",
                "min_time": "2018-02-26T00:00:00.000000000Z",
            },
            {
                "frequency": "4h",
                "max_time": "2022-07-12T12:00:00.000000000Z",
                "min_time": "2018-02-26T00:00:00.000000000Z",
            },
            {
                "frequency": "1d",
                "max_time": "2022-07-11T00:00:00.000000000Z",
                "min_time": "2018-02-26T00:00:00.000000000Z",
            },
        ],
        "pair": "abt-usd",
    },
]

catalog_metrics_test_data = [
    {
        "metric": "AdrActCnt",
        "full_name": "Addresses, active, count",
        "description": "The sum count of unique addresses that were active in the network (either as a recipient or originator of a ledger change) that interval. All parties in a ledger change action (recipients and originators) are counted. Individual addresses are not double-counted if previously active.",
        "category": "Addresses",
        "subcategory": "Active",
        "unit": "Addresses",
        "data_type": "bigint",
        "frequencies": [
            {"frequency": "1b", "assets": ["btc", "eth"]},
            {"frequency": "1d", "assets": ["ada", "btc", "eth"]},
        ],
    },
    {
        "metric": "FlowInBFXNtv",
        "full_name": "Flow, in, to Bitfinex, native units",
        "description": "The sum in native units sent to Bitfinex that interval.",
        "category": "Exchange",
        "subcategory": "Deposits",
        "unit": "Native units",
        "data_type": "decimal",
        "frequencies": [
            {"frequency": "1b", "assets": ["btc", "eth"]},
            {"frequency": "1d", "assets": ["btc", "eth"]},
        ],
        "reviewable": "True",
    },
]

catalog_markets_test_data = [
    {
        "market": "bitmex-XBTF15-future",
        "min_time": "2014-11-24T13:05:32.850000000Z",
        "max_time": "2015-01-30T12:00:00.000000000Z",
        "trades": {
            "min_time": "2014-11-24T13:05:32.850000000Z",
            "max_time": "2015-01-30T12:00:00.000000000Z",
        },
        "exchange": "bitmex",
        "type": "future",
        "symbol": "XBTF15",
        "base": "btc",
        "quote": "usd",
        "size_asset": "XBT",
        "margin_asset": "USD",
        "contract_size": "1",
        "tick_size": "0.1",
        "listing": "2014-11-24T13:05:32.850000000Z",
        "expiration": "2015-01-30T12:00:00.000000000Z",
    },
    {
        "market": "bitfinex-agi-btc-spot",
        "min_time": "2018-04-07T16:25:55.000000000Z",
        "max_time": "2020-03-25T20:12:09.639000000Z",
        "trades": {
            "min_time": "2018-04-07T16:25:55.000000000Z",
            "max_time": "2020-03-25T20:12:09.639000000Z",
        },
        "exchange": "bitfinex",
        "type": "spot",
        "base": "agi",
        "quote": "btc",
    },
]

catalog_exchanges_test_data = [
    {
        "exchange": "bibox",
        "markets": ["bibox-abt-btc-spot", "bibox-etc-usdt-spot"],
        "min_time": "2019-04-24T11:09:59.000000000Z",
        "max_time": "2019-05-18T16:06:10.927000000Z",
        "metrics": [
            {
                "metric": "volume_reported_spot_usd_1d",
                "frequencies": [
                    {
                        "frequency": "1d",
                        "min_time": "2019-04-30T00:00:00.000000000Z",
                        "max_time": "2019-06-16T00:00:00.000000000Z",
                    }
                ],
            },
            {
                "metric": "volume_reported_spot_usd_1h",
                "frequencies": [
                    {
                        "frequency": "1h",
                        "min_time": "2019-04-29T21:00:00.000000000Z",
                        "max_time": "2019-06-17T15:00:00.000000000Z",
                    }
                ],
            },
        ],
    },
    {
        "exchange": "binance",
        "markets": [
            "binance-BTCUSDT-future",
            "binance-LTCUSDT-future",
            "binance-ada-bnb-spot",
            "binance-btc-usdt-spot",
            "binance-bcpt-btc-spot",
            "binance-bcd-eth-spot",
        ],
        "min_time": "2017-07-14T04:00:00.510000000Z",
        "max_time": "2020-06-08T20:33:28.868000000Z",
        "metrics": [
            {
                "metric": "volume_reported_spot_usd_1d",
                "frequencies": [
                    {
                        "frequency": "1d",
                        "min_time": "2019-04-30T00:00:00.000000000Z",
                        "max_time": "2019-06-16T00:00:00.000000000Z",
                    }
                ],
            },
            {
                "metric": "volume_reported_spot_usd_1h",
                "frequencies": [
                    {
                        "frequency": "1h",
                        "min_time": "2019-04-29T21:00:00.000000000Z",
                        "max_time": "2019-06-17T15:00:00.000000000Z",
                    }
                ],
            },
        ],
    },
]

catalog_exchange_assets_test_data = [
    {
        "exchange_asset": "binance-btc",
        "metrics": [
            {
                "metric": "volume_trusted_spot_usd_1d",
                "frequencies": [
                    {
                        "frequency": "1d",
                        "min_time": "2020-10-16T00:00:00.000000000Z",
                        "max_time": "2021-01-05T00:00:00.000000000Z",
                    }
                ],
            },
            {
                "metric": "volume_trusted_spot_usd_1h",
                "frequencies": [
                    {
                        "frequency": "1h",
                        "min_time": "2020-10-15T03:00:00.000000000Z",
                        "max_time": "2021-01-06T12:00:00.000000000Z",
                    }
                ],
            },
        ],
    },
    {
        "exchange_asset": "coinbase-eth",
        "metrics": [
            {
                "metric": "volume_trusted_spot_usd_1d",
                "frequencies": [
                    {
                        "frequency": "1d",
                        "min_time": "2020-10-11T00:00:00.000000000Z",
                        "max_time": "2021-01-05T00:00:00.000000000Z",
                    }
                ],
            },
            {
                "metric": "volume_trusted_spot_usd_1h",
                "frequencies": [
                    {
                        "frequency": "1h",
                        "min_time": "2020-10-10T19:00:00.000000000Z",
                        "max_time": "2021-01-06T12:00:00.000000000Z",
                    }
                ],
            },
        ],
    },
]

catalog_asset_pairs_test_data = [
    {
        "pair": "aave-bnb",
        "metrics": [
            {
                "metric": "volume_trusted_spot_usd_1d",
                "frequencies": [
                    {
                        "frequency": "1d",
                        "min_time": "2020-10-16T00:00:00.000000000Z",
                        "max_time": "2021-01-05T00:00:00.000000000Z",
                    }
                ],
            },
            {
                "metric": "volume_trusted_spot_usd_1h",
                "frequencies": [
                    {
                        "frequency": "1h",
                        "min_time": "2020-10-15T03:00:00.000000000Z",
                        "max_time": "2021-01-06T12:00:00.000000000Z",
                    }
                ],
            },
        ],
    },
    {
        "pair": "aave-btc",
        "metrics": [
            {
                "metric": "volume_trusted_spot_usd_1d",
                "frequencies": [
                    {
                        "frequency": "1d",
                        "min_time": "2020-10-11T00:00:00.000000000Z",
                        "max_time": "2021-01-05T00:00:00.000000000Z",
                    }
                ],
            },
            {
                "metric": "volume_trusted_spot_usd_1h",
                "frequencies": [
                    {
                        "frequency": "1h",
                        "min_time": "2020-10-10T19:00:00.000000000Z",
                        "max_time": "2021-01-06T12:00:00.000000000Z",
                    }
                ],
            },
        ],
    },
]

catalog_institutions_test_data = [
    {
        "institution": "grayscale",
        "metrics": [
            {
                "metric": "gbtc_total_assets",
                "frequencies": [
                    {
                        "frequency": "1d",
                        "min_time": "2020-10-16T00:00:00.000000000Z",
                        "max_time": "2021-01-05T00:00:00.000000000Z",
                    }
                ],
            },
            {
                "metric": "gbtc_shares_outstanding",
                "frequencies": [
                    {
                        "frequency": "1h",
                        "min_time": "2020-10-15T03:00:00.000000000Z",
                        "max_time": "2021-01-06T12:00:00.000000000Z",
                    }
                ],
            },
        ],
    }
]

catalog_indexes_test_data = [
    {
        "index": "CMBI10",
        "description": "'CMBI10' index.",
        "frequencies": [
            {
                "frequency": "15s",
                "min_time": "2020-06-08T20:12:40.000000000Z",
                "max_time": "2020-06-08T20:29:30.000000000Z",
            }
        ],
    },
    {
        "index": "CMBIBTC",
        "description": "'CMBIBTC' index.",
        "frequencies": [
            {
                "frequency": "15s",
                "min_time": "2010-07-18T20:00:00.000000000Z",
                "max_time": "2020-06-08T20:29:45.000000000Z",
            },
            {
                "frequency": "1d",
                "min_time": "2010-07-19T00:00:00.000000000Z",
                "max_time": "2020-06-08T00:00:00.000000000Z",
            },
            {
                "frequency": "1d-ny-close",
                "min_time": "2010-07-18T20:00:00.000000000Z",
                "max_time": "2020-06-08T20:00:00.000000000Z",
            },
            {
                "frequency": "1d-sg-close",
                "min_time": "2010-07-19T08:00:00.000000000Z",
                "max_time": "2020-06-08T08:00:00.000000000Z",
            },
            {
                "frequency": "1h",
                "min_time": "2010-07-18T20:00:00.000000000Z",
                "max_time": "2020-06-08T20:00:00.000000000Z",
            },
        ],
    },
]

catalog_asset_alerts_test_data = [
    {
        "asset": "btc",
        "name": "block_count_empty_6b_hi",
        "description": "description",
        "threshold": "4",
        "constituents": ["block_count_empty_6b"],
    }
]

catalog_market_trades_data = [
    {
        "market": "binance-aave-btc-spot",
        "min_time": "2020-10-15T03:00:01.286000000Z",
        "max_time": "2022-07-12T21:00:39.933000000Z",
    },
    {
        "market": "binance-algo-usdc-spot",
        "min_time": "2019-06-22T00:00:07.600000000Z",
        "max_time": "2020-01-07T07:06:19.943000000Z",
    },
    {
        "market": "binance-atom-usdc-spot",
        "min_time": "2019-05-07T12:05:42.347000000Z",
        "max_time": "2022-07-12T21:01:25.281000000Z",
    },
]

catalog_market_quotes_data = [
    {
        "market": "coinbase-btc-usd-spot",
        "min_time": "2019-03-25T18:46:25.000000000Z",
        "max_time": "2022-07-12T21:01:40.000000000Z",
    },
    {
        "market": "coinbase-btc-usdt-spot",
        "min_time": "2021-08-20T13:00:00.000000000Z",
        "max_time": "2022-07-12T21:00:00.000000000Z",
    },
    {
        "market": "coinbase-eth-usd-spot",
        "min_time": "2019-03-25T18:46:25.000000000Z",
        "max_time": "2022-07-12T21:01:40.000000000Z",
    },
]

catalog_market_orderbook_data = [
    {
        "market": "binance-aave-btc-spot",
        "min_time": "2021-09-14T16:00:00.000000000Z",
        "max_time": "2022-07-12T20:00:00.000000000Z",
    },
    {
        "market": "binance-atom-usdc-spot",
        "min_time": "2021-09-14T16:00:00.000000000Z",
        "max_time": "2022-07-12T21:00:00.000000000Z",
    },
]


def get_empty_data(x: Any, y: Any) -> DataReturnType:
    return {"data": []}


def test_catalog_assets_request(mocker: Any) -> None:
    client = CoinMetricsClient("xxx")
    mock = Mock()
    mock.content = '{"data": [{"asset": "btc", "markets": ["coinbase-btc-usd-spot"]}]}'
    mocked_obj = mocker.patch.object(requests.Session, "get", return_value=mock)
    response = client.catalog_assets(assets="btc")
    mocked_obj.assert_called_once_with(
        "https://api.coinmetrics.io/v4/catalog/assets?api_key=xxx&assets=btc",
        verify=True,
        headers=client._session.headers,
        proxies={"http": None, "https": None},
    )
    assert response == [{"asset": "btc", "markets": ["coinbase-btc-usd-spot"]}]


def test_base_url() -> None:
    assert CoinMetricsClient("xxx")._api_base_url == "https://api.coinmetrics.io/v4"
    assert CoinMetricsClient("xxx")._api_key_url_str == "api_key=xxx"
    assert (
        CoinMetricsClient()._api_base_url == "https://community-api.coinmetrics.io/v4"
    )
    assert CoinMetricsClient()._api_key_url_str == ""


def test_to_dataframe() -> None:
    test_param_dict: Dict[str, UrlParamTypes] = {"asset": "btc", "metric": "1"}
    test_data_dict: DataReturnType = {"data": [test_param_dict for i in range(2)]}
    test_header = ["col1", "col2"]

    test_data_retrieval_function: DataRetrievalFuncType = lambda x, y: test_data_dict

    test_data_collection: DataCollection = DataCollection(
        data_retrieval_function=test_data_retrieval_function,
        endpoint="",
        url_params=test_param_dict,
    )
    test_df: DataFrameType = test_data_collection.to_dataframe()
    assert test_df.shape == (2, 2)
    assert list(test_param_dict.keys()) == list(test_df.columns)
    assert (test_df["asset"] == "btc").all()

    test_data_collection_header = DataCollection(
        data_retrieval_function=test_data_retrieval_function,
        endpoint="",
        url_params=test_param_dict,
    )
    test_df_header: DataFrameType = test_data_collection_header.to_dataframe(
        header=test_header
    )
    assert list(test_df_header.columns) == test_header


def test_catalog_assets_dataframe() -> None:
    data = CatalogAssetsData(catalog_assets_test_data)
    df = data.to_dataframe()
    assert len(df) == df.asset.nunique()

    df_markets = data.to_dataframe(secondary_level="markets")
    df_markets["metrics"] = df_markets["metrics"].astype(str)
    df_markets_test = pd.read_csv("test/data/catalog_assets_markets.csv", dtype=str)
    assert (df_markets.values == df_markets_test.values).all()

    df_metrics = data.to_dataframe(secondary_level="metrics")
    df_metrics = df_metrics.fillna("").astype(str)
    df_metrics_test = pd.read_csv(
        "test/data/catalog_assets_metrics.csv", dtype=str
    ).fillna("")
    assert (df_metrics_test.values == df_metrics.values).all()


def test_catalog_metrics_dataframe() -> None:
    data = CatalogMetricsData(catalog_metrics_test_data)
    df = data.to_dataframe().fillna("")
    df_test = pd.read_csv("test/data/catalog_metrics.csv", dtype=str).fillna("")
    assert (df.values == df_test.values).all()


def test_catalog_markets_dataframe() -> None:
    data = CatalogMarketsData(catalog_markets_test_data)
    df = data.to_dataframe().fillna("").astype(str)
    df_test = pd.read_csv("test/data/catalog_markets.csv", dtype=str).fillna("")
    assert (df.values == df_test.values).all()


def test_catalog_exchanges_dataframe() -> None:
    data = CatalogExchangesData(catalog_exchanges_test_data)
    df = data.to_dataframe()
    assert len(df) == df.exchange.nunique()
    df_markets = data.to_dataframe(secondary_level="markets")
    df_markets = df_markets.fillna("").astype(str)
    df_markets_test = pd.read_csv(
        "test/data/catalog_exchanges_markets.csv", dtype=str
    ).fillna("")
    assert (df_markets.values == df_markets_test.values).all()

    df_metrics = data.to_dataframe(secondary_level="metrics")
    df_metrics = df_metrics.fillna("").astype(str)
    df_metrics_test = pd.read_csv(
        "test/data/catalog_exchanges_metrics.csv", dtype=str
    ).fillna("")
    assert (df_metrics_test.values == df_metrics.values).all()


def test_catalog_exchange_assets_dataframe() -> None:
    data = CatalogExchangeAssetsData(catalog_exchange_assets_test_data)
    df = data.to_dataframe().fillna("").astype(str)
    df.to_csv("test/data/catalog_exchange_assets.csv", index_label=False)
    df_test = pd.read_csv("test/data/catalog_exchange_assets.csv", dtype=str).fillna("")
    assert (df.values == df_test.values).all()


def test_catalog_indexes_dataframe() -> None:
    data = CatalogIndexesData(catalog_indexes_test_data)
    df = data.to_dataframe().fillna("").astype(str)
    df_test = pd.read_csv("test/data/catalog_indexes.csv", dtype=str).fillna("")
    assert (df.values == df_test.values).all()


def test_catalog_institutions_dataframe() -> None:
    data = CatalogInstitutionsData(catalog_institutions_test_data)
    df = data.to_dataframe().fillna("").astype(str)
    print(os.getcwd())
    df_test = pd.read_csv("test/data/catalog_institutions.csv", dtype=str).fillna("")
    assert (df.values == df_test.values).all()


def test_catalog_asset_pairs_dataframe() -> None:
    data = CatalogAssetPairsData(catalog_asset_pairs_test_data)
    df = data.to_dataframe().fillna("").astype(str)
    df_test = pd.read_csv("test/data/catalog_asset_pairs.csv", dtype=str).fillna("")
    assert (df.values == df_test.values).all()


def test_catalog_asset_pairs_candles_dataframe() -> None:
    data = CatalogAssetPairCandlesData(catalog_candles_test_data)
    df = data.to_dataframe().fillna("").astype(str)
    df_test = pd.read_csv("test/data/catalog_asset_pair_candles.csv", dtype=str).fillna(
        ""
    )
    assert (df.values == df_test.values).all()


def test_catalog_market_trades_dataframe() -> None:
    data = CatalogMarketTradesData(catalog_market_trades_data)
    df = data.to_dataframe().fillna("").astype(str)
    df_test = pd.read_csv("test/data/catalog_market_trades.csv", dtype=str).fillna("")
    assert (df.values == df_test.values).all()


def test_catalog_market_orderbook_dataframe() -> None:
    data = CatalogMarketTradesData(catalog_market_orderbook_data)
    df = data.to_dataframe().fillna("").astype(str)
    df_test = pd.read_csv("test/data/catalog_market_orderbooks.csv", dtype=str).fillna(
        ""
    )
    assert (df.values == df_test.values).all()


def test_catalog_market_quotes() -> None:
    data = CatalogMarketTradesData(catalog_market_quotes_data)
    df = data.to_dataframe().fillna("").astype(str)
    df_test = pd.read_csv("test/data/catalog_market_quotes.csv", dtype=str).fillna("")
    assert (df.values == df_test.values).all()


def test_empty_dataframe() -> None:
    """Test client behavior when query results in no data"""
    empty_response = DataCollection(get_empty_data, "", {})
    assert empty_response
    assert empty_response.first_page() == []
    df = empty_response.to_dataframe()
    assert df.empty


def test_nested_dataframe() -> None:
    """Test the output for a dataframe with nested values"""
    nested_data: Dict[str, Any] = {
        "data": [
            {
                "time": "2020-05-01T22:00:00.000000000Z",
                "constituents": [
                    {"asset": "btc", "weight": "0.6"},
                    {"asset": "eth", "weight": "0.4"},
                ],
            }
        ]
    }

    def _get_test_data(x: Any, y: Any) -> DataReturnType:
        return nested_data

    test_data_collection = DataCollection(_get_test_data, "", {})
    df_nested = test_data_collection.to_dataframe()
    assert pd.api.types.is_datetime64tz_dtype(df_nested["time"])
    assert pd.api.types.is_string_dtype(df_nested["constituents"])
    assert pd.api.types.is_dict_like(df_nested["constituents"])
    assert pd.api.types.is_list_like(df_nested["constituents"])


def test_timeseries_dataframe() -> None:
    """Test the output dataframe of a timeseries response"""
    test_data: Dict[str, Any] = {
        "data": [
            {
                "time": "2022-01-01T00:00:00.000000000Z",
                "str_column": "str_value_1",
                "int_column": "1",
                "float_column": "1.5",
            },
            {
                "time": "2022-01-01T01:00:00.000000000Z",
                "str_column": "str_value_2",
                "int_column": "2",
                "float_column": "3.0",
            },
        ]
    }

    def _get_test_data(x: Any, y: Any) -> DataReturnType:
        return test_data

    test_data_collection = DataCollection(_get_test_data, "", {})
    df_test = test_data_collection.to_dataframe()
    assert pd.api.types.is_datetime64tz_dtype(df_test["time"])
    assert pd.api.types.is_integer_dtype(df_test["int_column"])
    assert pd.api.types.is_float_dtype(df_test["float_column"])
    assert pd.api.types.is_string_dtype(df_test["str_column"])


def test_export_to_csv() -> None:
    """Test the export_to_csv function"""
    test_data: Dict[str, Any] = {
        "data": [
            {
                "time": "2022-01-01T00:00:00.000000000Z",
                "test_column": "test_value1",
                "test_nested": [
                    {"asset": "btc", "weight": "0.6"},
                    {"asset": "eth", "weight": "0.4"},
                ],
            },
            {
                "time": "2022-01-01T01:00:00.000000000Z",
                "test_column": "test_value2",
                "test_nested": [
                    {"asset": "btc", "weight": "0.5"},
                    {"asset": "eth", "weight": "0.5"},
                ],
            },
        ]
    }

    def _get_test_data(x: Any, y: Any) -> DataReturnType:
        return test_data

    test_data_collection = DataCollection(_get_test_data, "", {})
    test_csv_str = test_data_collection.export_to_csv()
    if test_csv_str is not None:
        assert test_csv_str.split("\n")[0] == "time,test_column,test_nested"
        row1_str = test_csv_str.split("\n")[1]
        row1_value = row1_str.strip('"').split('","')
        assert len(row1_value) == len(test_data["data"][0].keys())
        assert row1_value[0] == test_data["data"][0]["time"]
        assert row1_value[1] == test_data["data"][0]["test_column"]
        assert row1_value[2] == str(test_data["data"][0]["test_nested"])
    else:
        raise



import os
import pprint

import typer
from datetime import datetime
from coinmetrics.data_exporter import CoinMetricsDataExporter
from typing import Union

coinmetrics_app = typer.Typer()
export_app = typer.Typer()

STRING_LIST_HELP_MSG = (
    "Pass in arguments as a list of strings separated by by commas i.e. {}"
)
HELP_MSG_EXCHANGES = STRING_LIST_HELP_MSG.format("binance,coinbase,bitmex")
HELP_MSG_ASSET_PAIRS = STRING_LIST_HELP_MSG.format("1INCHUSDT,BTCUSDT")
DATE_ARGUMENT = typer.Argument(str(datetime.now()).split(" ")[0])


@export_app.command()
def market_candles_spot(
    frequency: str,
    exchanges: str = typer.Argument(..., help=HELP_MSG_EXCHANGES),
    start_date: datetime = DATE_ARGUMENT,
    end_date: datetime = DATE_ARGUMENT,
    output_format: str = "json.gz",
    threaded: bool = False,
    api_key: Union[str, None] = None,
) -> None:
    """
    Used to export data related to spot market candles.
    Format: coinmetrics export market-candles-spot <exchanges> <start_date> <end_date>
    Example: coinmetrics export market-candles-spot coinbase,binance 2022-01-01 2022-01-03.
    """
    list_exchanges = exchanges.split(",")
    api_key = api_key or os.environ["CM_API_KEY"]
    data_exporter = CoinMetricsDataExporter(api_key=api_key)
    data_exporter.export_market_candles_spot_data(
        frequency=frequency,
        start_date=start_date,
        end_date=end_date,
        exchanges=list_exchanges,
        output_format=output_format,
        threaded=threaded,
    )


@export_app.command()
def market_candles_future(
    frequency: str,
    exchanges: str = typer.Argument(..., help=HELP_MSG_EXCHANGES),
    start_date: datetime = DATE_ARGUMENT,
    end_date: datetime = DATE_ARGUMENT,
    output_format: str = "json.gz",
    threaded: bool = False,
    api_key: Union[str, None] = None,
) -> None:
    """
    Used to export data related to future market candles.
    Format: coinmetrics export market-candles-future <exchanges> <start_date> <end_date>
    Example: coinmetrics export market-candles-future coinbase,binance 2022-01-01 2022-01-03.
    """
    list_exchanges = exchanges.split(",")
    api_key = api_key or os.environ["CM_API_KEY"]
    data_exporter = CoinMetricsDataExporter(api_key=api_key)
    data_exporter.export_market_candles_future_data(
        frequency=frequency,
        start_date=start_date,
        end_date=end_date,
        exchanges=list_exchanges,
        output_format=output_format,
        threaded=threaded,
    )


@export_app.command()
def market_trades_spot(
    exchanges: str = typer.Argument(..., help=HELP_MSG_EXCHANGES),
    start_date: datetime = DATE_ARGUMENT,
    end_date: datetime = DATE_ARGUMENT,
    output_format: str = "json.gz",
    threaded: bool = False,
    api_key: Union[str, None] = None,
) -> None:
    """
    Used to export data related to spot market trades.
    Format: coinmetrics export market-trades-spot <exchanges> <start_date> <end_date>
    Example: coinmetrics export market-trades-spot coinbase,binance 2022-01-01 2022-01-03.
    """
    list_exchanges = exchanges.split(",")
    api_key = api_key or os.environ["CM_API_KEY"]
    data_exporter = CoinMetricsDataExporter(api_key=api_key)
    data_exporter.export_market_trades_spot_data(
        start_date=start_date,
        end_date=end_date,
        exchanges=list_exchanges,
        output_format=output_format,
        threaded=threaded,
    )


@export_app.command()
def market_trades_future(
    exchanges: str = typer.Argument(..., help=HELP_MSG_EXCHANGES),
    start_date: datetime = DATE_ARGUMENT,
    end_date: datetime = DATE_ARGUMENT,
    output_format: str = "json.gz",
    threaded: bool = False,
    api_key: Union[str, None] = None,
) -> None:
    """
    Used to export data related to future market trades.
    Format: coinmetrics export market-trades-future <exchanges> <start_date> <end_date>
    Example: coinmetrics export market-trades-future coinbase,binance 2022-01-01 2022-01-03.
    """
    list_exchanges = exchanges.split(",")
    api_key = api_key or os.environ["CM_API_KEY"]
    data_exporter = CoinMetricsDataExporter(api_key=api_key)
    data_exporter.export_market_trades_future_data(
        start_date=start_date,
        end_date=end_date,
        exchanges=list_exchanges,
        output_format=output_format,
        threaded=threaded,
    )


@export_app.command()
def market_quotes_spot(
    exchanges: str = typer.Argument(..., help=HELP_MSG_EXCHANGES),
    asset_pairs: str = typer.Argument(..., help=HELP_MSG_ASSET_PAIRS),
    start_date: datetime = DATE_ARGUMENT,
    end_date: datetime = DATE_ARGUMENT,
    output_format: str = "json.gz",
    threaded: bool = False,
    api_key: Union[str, None] = None,
) -> None:
    """
    Used to export data related to spot market quotes
    Format: coinmetrics export market-quotes-spot <exchanges> <asset_pairs> <start_date> <end_date>
    Example: coinmetrics export market-quotes-spot binance 1INCHUSDT 2022-03-01 2022-03-03.
    """
    list_exchanges, list_asset_pairs = exchanges.split(","), asset_pairs.split(",")
    list_exchanges = [exchange.capitalize() for exchange in exchanges]
    api_key = api_key or os.environ["CM_API_KEY"]
    data_exporter = CoinMetricsDataExporter(api_key=api_key)
    data_exporter.export_market_quotes_spot_data(
        start_date=start_date,
        end_date=end_date,
        exchanges=list_exchanges,
        asset_pairs=list_asset_pairs,
        output_format=output_format,
        threaded=threaded,
    )


@export_app.command()
def market_quotes_future(
    exchanges: str = typer.Argument(..., help=HELP_MSG_EXCHANGES),
    asset_pairs: str = typer.Argument(..., help=HELP_MSG_ASSET_PAIRS),
    start_date: datetime = DATE_ARGUMENT,
    end_date: datetime = DATE_ARGUMENT,
    output_format: str = "json.gz",
    threaded: bool = False,
    api_key: Union[str, None] = None,
) -> None:
    """
    Used to export data related to future market quotes.
    Format: coinmetrics export market-quotes-future <exchanges> <asset_pairs> <start_date> <end_date>
    Example: coinmetrics export market-quotes-future binance 1INCHUSDT 2022-03-01 2022-03-03.
    """
    list_exchanges, list_asset_pairs = exchanges.split(","), asset_pairs.split(",")
    list_exchanges = [exchange.capitalize() for exchange in exchanges]
    api_key = api_key or os.environ["CM_API_KEY"]
    data_exporter = CoinMetricsDataExporter(api_key=api_key)
    data_exporter.export_market_quotes_future_data(
        start_date=start_date,
        end_date=end_date,
        exchanges=list_exchanges,
        asset_pairs=list_asset_pairs,
        output_format=output_format,
        threaded=threaded,
    )


@export_app.command()
def get_exchanges(
    command: str = typer.Argument(...), api_key: Union[str, None] = None
) -> None:
    api_key = api_key or os.environ["CM_API_KEY"]
    data_exporter = CoinMetricsDataExporter(api_key=api_key)
    pprint.pprint(data_exporter.get_exchanges(command))


@export_app.command()
def get_asset_pairs(
    file_type: str = typer.Argument(...),
    exchange: str = typer.Argument(...),
    api_key: Union[str, None] = None,
) -> None:
    exchange = exchange.capitalize()
    api_key = api_key or os.environ["CM_API_KEY"]
    data_exporter = CoinMetricsDataExporter(api_key=api_key)
    pprint.pprint(data_exporter.get_asset_pairs(file_type=file_type, exchange=exchange))


EXPORT_APP_HELP_MSG = """
This app is used to programatically download files from the CoinMetrics flat files server. It supports exporting
daily data files related to market-trades, market-quotes, and market candles in json.gz, csv, or json formats.
To specify multiple exchanges or multiple asset_pairs use comma seperation in order to pass in a list.
Examples:

To download binance spot trade files for month of June 2019:
coinmetrics export market-trades-spot binance 2019-06-01 2019-06-30 --output-format csv

To download binance and coinbase market-trade-spot files for 10 days:
 coinmetrics export market-trades-spot binance,coinbase 2018-01-01 2018-01-10

To download market-quotes-futures file for binance 1INCHUSDT pair for only one day:
 coinmetrics export market-quotes-future binance 1INCHUSDT 2022-02-19 2022-02-19

To run any of the commands "threaded" or concurrently add the flag --threaded i.e.
 coinmetrics export market-trades-spot binance 2019-06-01 2019-06-30 --output-format csv --threaded
Note this may increase the speed of download but will use more system resources and bandwith.
"""


def main() -> None:
    coinmetrics_app.add_typer(export_app, name="export", help=EXPORT_APP_HELP_MSG)
    coinmetrics_app()


if __name__ == "__main__":
    main()



import csv
import logging
import orjson
import requests
import socket
import os
import zlib
import sys
from requests import Response, HTTPError
from requests.auth import HTTPBasicAuth

from coinmetrics._exceptions import CoinMetricsUnauthorizedException
from coinmetrics._utils import retry
from datetime import datetime
from typing import Union, Optional, List, Dict, Any, Iterable
from concurrent.futures import ThreadPoolExecutor

DEFAULT_OUTPUT_FORMAT = "json.gz"
Log_Format = "%(levelname)s %(asctime)s - %(message)s"

logging.basicConfig(
    stream=sys.stdout, filemode="w", format=Log_Format, level=logging.INFO
)

logger = logging.getLogger()


class CoinMetricsDataExporter:
    """
    Class to facilitate exporting flat files for coinmetrics market data to various data formats
    """

    _API_BASE_URL = "https://files.coinmetrics.io/api/"
    _SUPPORTED_FREQUENCIES = ["1m", "5m", "10m", "15m", "30m", "1h", "1d"]

    def __init__(
        self,
        api_key: str = "",
        verify_ssl_certs: Union[bool, str] = True,
        proxy_url: Optional[str] = None,
        session: Optional[requests.Session] = None,
    ):
        self._api_key = api_key
        self._verify_ssl_certs = verify_ssl_certs
        self._proxy_url = proxy_url
        self._auth = HTTPBasicAuth(self._api_key, "")
        if session is None:
            self._session = requests.Session()
            self._session.verify = self._verify_ssl_certs
        else:
            self._session = session

    def export_market_candles_future_data(
        self,
        frequency: str,
        start_date: datetime,
        end_date: datetime,
        exchanges: Union[str, List[str]],
        output_format: str = DEFAULT_OUTPUT_FORMAT,
        threaded: bool = False,
    ) -> None:
        if frequency not in self._SUPPORTED_FREQUENCIES:
            raise ValueError(
                f"This frequency is not supported for market-candles-futures,"
                f" supported frequencies: {self._SUPPORTED_FREQUENCIES}"
            )
        candles_future_route = f"market-candles-future-{frequency}"
        list_files_to_download = self._get_list_of_file_urls_to_download_market_trades(
            start_date=start_date,
            end_date=end_date,
            route_url=candles_future_route,
            exchanges=exchanges,
        )
        self._download_list_of_files(
            list_files_to_download, output_format=output_format, threaded=threaded
        )

    def export_market_candles_spot_data(
        self,
        frequency: str,
        start_date: datetime,
        end_date: datetime,
        exchanges: Union[str, List[str]],
        output_format: str = DEFAULT_OUTPUT_FORMAT,
        threaded: bool = False,
    ) -> None:
        if frequency not in self._SUPPORTED_FREQUENCIES:
            raise ValueError(
                f"This frequency is not supported for market-candles-spot,"
                f" supported frequencies: {self._SUPPORTED_FREQUENCIES}"
            )
        candles_future_route = f"market-candles-spot-{frequency}"
        list_files_to_download = self._get_list_of_file_urls_to_download_market_trades(
            start_date=start_date,
            end_date=end_date,
            route_url=candles_future_route,
            exchanges=exchanges,
        )
        self._download_list_of_files(
            list_files_to_download, output_format=output_format, threaded=threaded
        )

    def export_market_quotes_future_data(
        self,
        start_date: datetime,
        end_date: datetime,
        asset_pairs: List[str],
        exchanges: Union[str, List[str]],
        output_format: str = DEFAULT_OUTPUT_FORMAT,
        threaded: bool = False,
    ) -> None:
        """
        Exports market quotes futures data to flat files given the parameters
        :param asset_pairs: list of string of asset pairs to download
        :param exchanges: list of string of exchanges
        :param start_date: date to start querying files
        :param end_date: date to end querying
        :param output_format: str how the files should be saved, supported data types are "json.gz", "json", and "csv"
        :param threaded: if true will thread the download process
        """
        if isinstance(exchanges, str):
            exchanges = exchanges.capitalize()
        else:
            exchanges = [exchange.capitalize() for exchange in exchanges]
        quotes_route = "market-quotes-future"
        list_files_to_download = self._get_list_files_to_download_from_ff_server(
            assets_pairs=asset_pairs,
            route_url=quotes_route,
            start_date=start_date,
            end_date=end_date,
            exchanges=exchanges,
        )
        self._download_list_of_files(
            list_files_to_download, output_format=output_format, threaded=threaded
        )

    def export_market_quotes_spot_data(
        self,
        start_date: datetime,
        end_date: datetime,
        asset_pairs: List[str],
        exchanges: Union[str, List[str]],
        output_format: str = DEFAULT_OUTPUT_FORMAT,
        threaded: bool = False,
    ) -> None:
        """
        Exports market quotes data to flat files given the parameters
        :param asset_pairs: list of string of asset pairs to download
        :param exchanges: list of string of exchanges
        :param start_date: date to start querying files
        :param end_date: date to end querying
        :param output_format: str how the files should be saved, supported data types are "json.gz", "json", and "csv"
        :param threaded: if true will thread the download process
        """
        if isinstance(exchanges, str):
            exchanges = exchanges.capitalize()
        else:
            exchanges = [exchange.capitalize() for exchange in exchanges]
        quotes_route = "market-quotes-spot"
        list_files_to_download = self._get_list_files_to_download_from_ff_server(
            assets_pairs=asset_pairs,
            route_url=quotes_route,
            start_date=start_date,
            end_date=end_date,
            exchanges=exchanges,
        )
        self._download_list_of_files(
            list_files_to_download, output_format=output_format, threaded=threaded
        )

    def export_market_trades_spot_data(
        self,
        start_date: datetime,
        end_date: datetime,
        exchanges: Union[str, List[str]],
        output_format: str = DEFAULT_OUTPUT_FORMAT,
        threaded: bool = False,
    ) -> None:
        """
        Exports market trade data to flat files given the parameters
        :param exchanges: list of string of exchanges
        :param start_date: date to start querying files
        :param end_date: date to end querying
        :param output_format: str how the files should be saved, supported data types are "json.gz", "json", and "csv"
        """
        market_trades_url = "market-trades-spot"
        list_files_to_download = self._get_list_of_file_urls_to_download_market_trades(
            start_date=start_date,
            end_date=end_date,
            exchanges=exchanges,
            route_url=market_trades_url,
        )
        self._download_list_of_files(
            list_files_to_download, output_format=output_format, threaded=threaded
        )

    def export_market_trades_future_data(
        self,
        start_date: datetime,
        end_date: datetime,
        exchanges: Union[str, List[str]],
        output_format: str = DEFAULT_OUTPUT_FORMAT,
        threaded: bool = False,
    ) -> None:
        """
        Exports market trade data to flat files given the parameters
        :param exchanges: list of string of exchanges
        :param start_date: date to start querying files
        :param end_date: date to end querying
        :param output_format: str how the files should be saved, supported data types are "json.gz", "json", and "csv"
        """
        market_trades_url = "market-trades-future"
        list_files_to_download = self._get_list_of_file_urls_to_download_market_trades(
            start_date=start_date,
            end_date=end_date,
            exchanges=exchanges,
            route_url=market_trades_url,
        )
        self._download_list_of_files(
            list_of_files=list_files_to_download,
            output_format=output_format,
            threaded=threaded,
        )

    def get_exchanges(self, file_type: str) -> List[str]:
        """
        Get the exchanges available for a file type
        :param file_type: str one of the queryable file types, such as "market-trades-future/spot",
        "market-candles-future/spot-<frequency>", "market-quotes-future/spot"
        """
        request_data = self._send_request(self._API_BASE_URL + file_type).json()
        result = [directory["name"] for directory in request_data]
        return result

    def get_asset_pairs(self, file_type: str, exchange: str) -> List[str]:
        request_data = self._send_request(
            f"{self._API_BASE_URL}/{file_type}/{exchange}"
        ).json()
        result = [directory["name"] for directory in request_data]
        return result

    def _download_list_of_files(
        self, list_of_files: List[str], output_format: str, threaded: bool = False
    ) -> None:
        """
        Downloads a list of files from the CM API server, gives the option to thread it or not
        :param list_of_files: list of files to download from server
        :output_format: str supported formats are json.gz, csv, or json
        :param threaded: if true will run download processes concurrently, which can make long running tasks run quicker.
        Downloading is still limited by bandwith
        """
        if threaded:
            with ThreadPoolExecutor() as executor:
                for file in list_of_files:
                    executor.submit(
                        self._download_file_from_server,
                        url=file,
                        output_format=output_format,
                    )
        else:
            for file in list_of_files:
                self._download_file_from_server(url=file, output_format=output_format)

    def _download_file_from_server(
        self, url: str, output_format: str = DEFAULT_OUTPUT_FORMAT
    ) -> None:
        """
        Downloads file from flat files server, converts the data type if neccesary, and stores it locally
        :param url: url to the file to download i.e. 'market-trades-spot/binance/2019-09-09.json.gz'
        :param output_format: str output format supported formats are 'csv', 'json', and 'json.gz'
        """
        url_split = url.split("/")
        file_name = url_split[-1]
        rest_of_url = "/".join(url_split[:-1])
        os.makedirs(rest_of_url, exist_ok=True)

        if output_format == "json.gz":
            with self._send_request(self._API_BASE_URL + url, stream=True) as response:
                with open(url, "wb") as file:
                    for chunk in response.iter_content(chunk_size=1024):
                        file.write(chunk)
                    logger.info(f"Downloaded file {url}")
            return

        elif output_format == "csv":
            file_name = file_name.replace("json.gz", "csv")
            full_file_name = f"{rest_of_url}/{file_name}"
            self._download_and_stream_data_to_csv(
                self._API_BASE_URL + url, full_file_name
            )

        elif output_format == "json":
            file_name = file_name.replace("json.gz", "json")
            full_file_name = f"{rest_of_url}/{file_name}"
            self._download_and_stream_data_to_json(
                request_url=self._API_BASE_URL + url, file_name=full_file_name
            )

        else:
            raise ValueError(f"Unsupported output type: {output_format}")

    def _get_list_files_to_download_from_ff_server(
        self,
        assets_pairs: List[str],
        exchanges: Union[str, List[str]],
        route_url: str,
        start_date: datetime,
        end_date: datetime,
    ) -> List[str]:
        """
        Queries the CoinMetrics flat files server and figures out what files to download
        """
        if isinstance(exchanges, str):
            exchanges = [exchanges]

        market_quotes_spot = self._API_BASE_URL + route_url
        market_quotes_exchange_folders = self._send_request(market_quotes_spot).json()
        list_of_files_to_download = []
        list_supported_exchanges = set()
        for market_trade_dir in market_quotes_exchange_folders:
            exchange_name = market_trade_dir["name"]
            list_supported_exchanges.add(exchange_name)

        for exchange in exchanges:
            if exchange not in list_supported_exchanges:
                raise ValueError(
                    f"Provided exchange: {exchange} is not supported for flat files. Supported exchanges: {list_supported_exchanges}"
                )

        for exchange in exchanges:
            exchange_files_url = market_quotes_spot + f"/{exchange}"
            exchange_asset_pair_data = self._send_request(exchange_files_url).json()
            exchange_asset_pair_folders = {
                folder["name"] for folder in exchange_asset_pair_data
            }

            asset_pair_url_dict = {}
            for asset_pair in assets_pairs:
                if asset_pair not in exchange_asset_pair_folders:
                    logger.info(
                        f"Asset pair: {asset_pair} not supported for exchange: {exchange}, skipping"
                    )
                else:
                    exchange_asset_pair_url = f"{exchange_files_url}/{asset_pair}"
                    asset_pair_url_dict[asset_pair] = exchange_asset_pair_url

            for asset_pair, asset_pair_url in asset_pair_url_dict.items():
                date_files_exchange_asset_pair = self._send_request(
                    asset_pair_url
                ).json()
                filtered_date_files = self.filter_date_files(
                    date_files_exchange_asset_pair,
                    start_date,
                    end_date,
                    f"{route_url}/{exchange}/{asset_pair}",
                )
                list_of_files_to_download.extend(filtered_date_files)

        return list_of_files_to_download

    def _get_list_of_file_urls_to_download_market_trades(
        self,
        start_date: datetime,
        end_date: datetime,
        exchanges: Union[str, List[str]],
        route_url: str,
    ) -> List[str]:
        if type(exchanges) == str:
            exchanges = [str(exchanges)]

        market_trades_url = self._API_BASE_URL + route_url
        market_trades_exchanges = self._send_request(market_trades_url).json()
        list_of_files_to_download = []
        list_supported_exchanges = []
        for market_trade_dir in market_trades_exchanges:
            exchange_name = market_trade_dir["name"]
            list_supported_exchanges.append(exchange_name)

        for exchange in exchanges:
            if exchange not in list_supported_exchanges:
                raise ValueError(
                    f"Provided exchange: {exchange} is not supported for flat files. Supported exchanges: {list_supported_exchanges}"
                )

            exchange_files_url = market_trades_url + f"/{exchange}"
            files = self._send_request(exchange_files_url).json()
            filtered_exchange_files = self.filter_date_files(
                files, start_date, end_date, f"{route_url}/{exchange}"
            )
            list_of_files_to_download.extend(filtered_exchange_files)

        return list_of_files_to_download

    def _download_and_stream_data_to_json(
        self, request_url: str, file_name: str
    ) -> None:
        """
        Streams gzipped data from url and decodes it to a valid json file
        :param request_url: str request url where json.gz file is stored
        :param file_name: name of the file to store
        """
        with self._send_request(request_url, stream=True) as resp:
            decoded = stream_gzip_decompress(resp.iter_content(chunk_size=1024))
            with open(file_name, "w") as file:
                file.write("[")
                for line in decoded:
                    full_line = line.decode("UTF-8").replace("}", "},")
                    file.write(full_line)
                # This is done in order to remove trailing ',' and add a ']' to conform with JSON spec
                file.seek(file.tell() - 1, os.SEEK_SET)
                file.write("")
                file.seek(file.tell() - 1, os.SEEK_SET)
                file.write("")
                file.write("]")

    def _download_and_stream_data_to_csv(
        self, request_url: str, file_name: str
    ) -> None:
        """
        Streams gzipped data from url and decodes it into a valid csv file
        :param request_url: str request url where json.gz file is stored
        :param file_name: name of the file to store
        """
        with self._send_request(request_url, stream=True) as resp:
            decoded_dicts = stream_gzip_decompress_to_dicts(resp.iter_content(1024))
            with open(file_name, "w") as csvfile:
                first_iteration = True
                for data in decoded_dicts:
                    if first_iteration:
                        writer = csv.DictWriter(
                            csvfile,
                            fieldnames=list(data[0].keys()),
                            extrasaction="ignore",
                        )
                        writer.writeheader()
                        first_iteration = False
                    writer.writerows(data)
                logger.info(f"Downloaded file {file_name}")

    @staticmethod
    def filter_date_files(
        list_of_dated_file_names: List[Dict[str, str]],
        start_date: datetime,
        end_date: datetime,
        result_file_abbreviation: str,
    ) -> List[str]:
        """
        Filters file names formatted like YYYY-MM-DD.<file_format> so that only files with dates that fall in between
         the start and end dates are returned
         :param list_of_dated_file_names: list of files names formatted as YYYY-MM-DD.<file_format>
         :param start_date: datetime start date of filter
         :param end_date: datetime end date of filter
         :param result_file_abbreviation: str abbreviation to append to front of file names. I.e.
          market-trades-spot/binance     would be included with all results
         :return: list of str file names that fall in between the start and end dates
        """
        result_files = []
        for file in list_of_dated_file_names:
            file_name = file["name"]
            file_date_str = file_name.split(".")[0]
            file_date = datetime.strptime(file_date_str, "%Y-%m-%d")
            if start_date <= file_date <= end_date:
                full_file_name = f"{result_file_abbreviation}/{file_name}"
                result_files.append(full_file_name)
        return result_files

    @retry((socket.gaierror, HTTPError), retries=5, wait_time_between_retries=5)
    def _send_request(self, actual_url: str, stream: bool = False) -> Response:
        response = self._session.get(
            actual_url, verify=self._verify_ssl_certs, auth=self._auth, stream=stream
        )
        if response.status_code == 403 or response.status_code == 401:
            raise CoinMetricsUnauthorizedException(response=response)
        if response.status_code != 200:
            response.raise_for_status()
        return response


def stream_gzip_decompress_to_dicts(
    stream: Iterable[bytes],
) -> Iterable[Dict[Any, Any]]:
    dec = zlib.decompressobj(32 + zlib.MAX_WBITS)
    remainder = ""
    for chunk in stream:
        rv = dec.decompress(chunk)
        if rv:
            test_line = rv.decode("UTF-8")
            line_split = test_line.split("}")
            full_dict = remainder + "},".join(line_split[:-1]) + "}"
            remainder = line_split[-1:][0]
            yield orjson.loads("[" + full_dict + "]")


def stream_gzip_decompress(stream: Iterable[bytes]) -> Iterable[bytes]:
    dec = zlib.decompressobj(32 + zlib.MAX_WBITS)
    for chunk in stream:
        rv = dec.decompress(chunk)
        if rv:
            yield rv



from dateutil.parser import isoparse
from typing import Iterable
from coinmetrics._typing import DataFrameType, List, Any, Optional
from logging import getLogger

logger = getLogger("cm_client")

try:
    import pandas as pd  # type: ignore
except ImportError:
    pd = None
    logger.warning(
        "Pandas export is unavailable. Install pandas to unlock dataframe functions."
    )


def _convert_utc(x: Any) -> DataFrameType:
    try:
        return isoparse(x)
    except TypeError:
        return None


def _expand_df(key: str, iterable: Iterable[Any]) -> List[Any]:
    def _assign_value(row: Any) -> Any:
        try:
            return row[key]
        except (KeyError, TypeError):
            return None

    return list(map(_assign_value, iterable))


def convert_catalog_dtypes(df: DataFrameType) -> DataFrameType:
    df = df.convert_dtypes()
    columns = df.columns
    datetime_cols = [c for c in columns if c.endswith("_time") or c == "time"]
    for col in datetime_cols:
        df[col] = df[col].apply(_convert_utc)
    return df


class CatalogAssetsData(List[Any]):
    def to_dataframe(self, secondary_level: Optional[str] = None) -> DataFrameType:
        """
        Transforms catalog data in list form into a dataframe

        :param secondary_level: Second level of aggregation next to exchanges. One of "markets" or "metrics"; raises ValueError if neither.
        :type secondary_level: str
        :return: Catalog Data
        """
        df_assets = pd.DataFrame(self)
        if secondary_level is None:
            pass
        elif secondary_level == "markets":
            df_assets = (
                df_assets.explode("markets")
                .fillna({"markets": ""})
                .rename({"markets": "market"}, axis=1)
                .assign(
                    exchange=lambda x: x["market"].apply(lambda row: row.split("-")[0])
                )
                .drop("exchanges", axis=1)
                .reset_index(drop=True)
            )

        elif secondary_level == "metrics":
            asset_mapper = df_assets.asset.to_dict()

            def _assign_metric(x: Any) -> Any:
                try:
                    return x["metric"]
                except TypeError:
                    return None

            df_assets = df_assets.explode("metrics").assign(
                metrics=lambda x: pd.Series(x["metrics"])
            )
            df_assets["metric"] = df_assets["metrics"].apply(_assign_metric)
            df_assets_metrics = df_assets.dropna(subset=["metrics"]).metrics.apply(
                pd.Series
            )
            df_assets_metrics["asset"] = df_assets_metrics.index.map(asset_mapper)
            df_assets_metrics = (
                df_assets_metrics.explode("frequencies")
                .assign(
                    frequency=lambda df: _expand_df(
                        key="frequency", iterable=df.frequencies
                    )
                )
                .assign(
                    min_time=lambda df: _expand_df(
                        key="min_time", iterable=df.frequencies
                    )
                )
                .assign(
                    max_time=lambda df: _expand_df(
                        key="max_time", iterable=df.frequencies
                    )
                )
                .assign(
                    min_height=lambda df: _expand_df(
                        key="min_height", iterable=df.frequencies
                    )
                )
                .assign(
                    max_height=lambda df: _expand_df(
                        key="max_height", iterable=df.frequencies
                    )
                )
                .assign(
                    min_hash=lambda df: _expand_df(
                        key="min_hash", iterable=df.frequencies
                    )
                )
                .assign(
                    max_hash=lambda df: _expand_df(
                        key="max_hash", iterable=df.frequencies
                    )
                )
                .drop(["frequencies"], axis=1)
            )
            df_assets = (
                df_assets.drop(["metrics"], axis=1)
                .merge(df_assets_metrics, on=["metric", "asset"], how="left")
                .reset_index(drop=True)
            )

        else:
            raise ValueError(
                "secondary_level must be one of None, 'markets' or 'metrics"
            )
        df_assets = convert_catalog_dtypes(df_assets)
        return df_assets


class CatalogAssetAlertsData(List[Any]):
    def to_dataframe(self) -> DataFrameType:
        """
        Transforms catalog data in list form into a dataframe
        :return: Catalog Data
        """
        return convert_catalog_dtypes(pd.DataFrame(self))


class CatalogMarketTradesData(List[Any]):
    def to_dataframe(self) -> DataFrameType:
        """
        Transforms catalog data in list form into a dataframe
        :return: Catalog Data
        """
        return convert_catalog_dtypes(pd.DataFrame(self))


class CatalogMarketOrderbooksData(List[Any]):
    def to_dataframe(self, secondary_level: Optional[str] = None) -> DataFrameType:
        """
        Transforms catalog data in list form into a dataframe
        :return: Catalog Data
        """
        df_assets = pd.DataFrame(self)
        if secondary_level is None:
            pass
        elif secondary_level == "depths":

            def _assign(x: Any, key: str = "depth") -> Any:
                try:
                    return x[key]
                except TypeError:
                    return None

            df_assets = df_assets.explode("depths").assign(
                depth=lambda x: pd.Series(x["depths"])
            )
            df_assets["depth"] = df_assets["depths"].apply(_assign, key="depth")
            df_assets["min_time"] = df_assets["depths"].apply(_assign, key="min_time")
            df_assets["max_time"] = df_assets["depths"].apply(_assign, key="max_time")
            df_assets.drop(["depths"], inplace=True, axis=1)
        else:
            raise ValueError

        return convert_catalog_dtypes(df_assets)


class CatalogAssetPairsData(List[Any]):
    def to_dataframe(self) -> DataFrameType:
        """
        Transforms catalog data in list form into a dataframe
        :return: Catalog Data
        """
        df_asset_pairs = (
            pd.DataFrame(self)
            .explode("metrics")
            .assign(metric=lambda df: _expand_df(key="metric", iterable=df.metrics))
            .assign(
                frequencies=lambda df: _expand_df(
                    key="frequencies", iterable=df.metrics
                )
            )
            .explode("frequencies")
            .assign(
                frequency=lambda df: _expand_df(
                    key="frequency", iterable=df.frequencies
                )
            )
            .assign(
                min_time=lambda df: _expand_df(key="min_time", iterable=df.frequencies)
            )
            .assign(
                max_time=lambda df: _expand_df(key="max_time", iterable=df.frequencies)
            )
            .reset_index(drop=True)
            .drop(["metrics", "frequencies"], axis=1)
        )
        return convert_catalog_dtypes(df_asset_pairs)


class CatalogExchangesData(List[Any]):
    def to_dataframe(self, secondary_level: Optional[str] = None) -> DataFrameType:
        """
        Transforms catalog data in list form into a dataframe

        :param secondary_level: Second level of aggregation next to exchanges. One of "markets" or "metrics"; raises ValueError if neither.
        :type secondary_level: str
        :return: Catalog Data
        """
        df_exchanges = pd.DataFrame(self)
        if secondary_level is None:
            pass
        elif secondary_level == "markets":
            df_exchanges = df_exchanges.explode("markets").rename(
                {"markets": "market"}, axis=1
            )
        elif secondary_level == "metrics":
            df_exchanges = (
                df_exchanges.explode("metrics")
                .assign(metric=lambda df: _expand_df(key="metric", iterable=df.metrics))
                .assign(
                    frequencies=lambda df: _expand_df(
                        key="frequencies", iterable=df.metrics
                    )
                )
                .explode("frequencies")
                .assign(
                    frequency=lambda df: _expand_df(
                        key="frequency", iterable=df.frequencies
                    )
                )
                .assign(
                    min_time=lambda df: _expand_df(
                        key="min_time", iterable=df.frequencies
                    )
                )
                .assign(
                    max_time=lambda df: _expand_df(
                        key="max_time", iterable=df.frequencies
                    )
                )
                .reset_index(drop=True)
                .drop(["metrics", "frequencies"], axis=1)
            )
        else:
            raise ValueError("secondary_level must be one of 'markets' or 'metrics")
        df_exchanges = convert_catalog_dtypes(df_exchanges)
        return df_exchanges


class CatalogExchangeAssetsData(List[Any]):
    def to_dataframe(self) -> DataFrameType:
        """
        Transforms catalog data in list form into a dataframe
        :return: Catalog Data
        """
        df_exchange_assets = pd.DataFrame(self)
        df_exchange_assets = (
            df_exchange_assets.explode("metrics")
            .assign(metric=lambda df: _expand_df(key="metric", iterable=df.metrics))
            .assign(
                frequencies=lambda df: _expand_df(
                    key="frequencies", iterable=df.metrics
                )
            )
            .explode("frequencies")
            .assign(
                frequency=lambda df: _expand_df(
                    key="frequency", iterable=df.frequencies
                )
            )
            .assign(
                min_time=lambda df: _expand_df(key="min_time", iterable=df.frequencies)
            )
            .assign(
                max_time=lambda df: _expand_df(key="max_time", iterable=df.frequencies)
            )
            .reset_index(drop=True)
            .drop(["metrics", "frequencies"], axis=1)
        )
        return convert_catalog_dtypes(df_exchange_assets)


class CatalogIndexesData(List[Any]):
    """
    Transforms catalog data in list form into a dataframe
    :return: Catalog Data
    """

    def to_dataframe(self) -> DataFrameType:
        df_indexes = pd.DataFrame(self)
        df_indexes = (
            df_indexes.explode("frequencies")
            .assign(
                frequency=lambda df: _expand_df(
                    key="frequency", iterable=df.frequencies
                )
            )
            .assign(
                min_time=lambda df: _expand_df(key="min_time", iterable=df.frequencies)
            )
            .assign(
                max_time=lambda df: _expand_df(key="max_time", iterable=df.frequencies)
            )
            .reset_index(drop=True)
            .drop(["frequencies"], axis=1)
        )
        return convert_catalog_dtypes(df_indexes)


class CatalogInstitutionsData(List[Any]):
    """
    Transforms catalog data in list form into a dataframe
    :return: Catalog Data
    """

    def to_dataframe(self) -> DataFrameType:
        df_institutions = pd.DataFrame(self)
        df_institutions = (
            df_institutions.explode("metrics")
            .assign(metric=lambda df: _expand_df(key="metric", iterable=df.metrics))
            .assign(
                frequencies=lambda df: _expand_df(
                    key="frequencies", iterable=df.metrics
                )
            )
            .explode("frequencies")
            .assign(
                frequency=lambda df: _expand_df(
                    key="frequency", iterable=df.frequencies
                )
            )
            .assign(
                min_time=lambda df: _expand_df(key="min_time", iterable=df.frequencies)
            )
            .assign(
                max_time=lambda df: _expand_df(key="max_time", iterable=df.frequencies)
            )
            .reset_index(drop=True)
            .drop(["metrics", "frequencies"], axis=1)
        )
        return convert_catalog_dtypes(df_institutions)


class CatalogMarketsData(List[Any]):
    """
    Transforms catalog data in list form into a dataframe
    :return: Catalog Data
    """

    def to_dataframe(self) -> DataFrameType:
        df_markets = pd.DataFrame(self)
        df_markets = df_markets.join(
            df_markets["trades"].apply(pd.Series).drop(0, axis=1, errors="ignore"),
            rsuffix="_trades",
        )
        metadata = ["funding_rates", "openinterest", "liquidations"]
        for col in metadata:
            if col in df_markets.columns:
                df_markets = df_markets.join(
                    df_markets[col].apply(pd.Series).drop(0, axis=1, errors="ignore"),
                    rsuffix=f"_{col}",
                )

        df_markets = df_markets.drop(
            ["trades", "funding_rates", "openinterest", "liquidations"],
            axis=1,
            errors="ignore",
        )
        return convert_catalog_dtypes(df_markets)


class CatalogMetricsData(List[Any]):
    """
    Transforms catalog data in list form into a dataframe
    :return: Catalog Data
    """

    def to_dataframe(self) -> DataFrameType:
        df_catalog_metrics = (
            pd.DataFrame(self)
            .explode("frequencies")
            .assign(
                frequency=lambda df: _expand_df(
                    key="frequency", iterable=df.frequencies
                )
            )
            .assign(asset=lambda df: _expand_df(key="assets", iterable=df.frequencies))
            .explode("asset")
            .drop("frequencies", axis=1)
            .reset_index(drop=True)
        )
        return convert_catalog_dtypes(df_catalog_metrics)


class CatalogExchangeAssetMetricsData(List[Any]):
    """
    Transforms catalog exchange asset data in list form into a dataframe
    :return: Catalog Data
    """

    def to_dataframe(self) -> DataFrameType:
        df_catalog_metrics = (
            pd.DataFrame(self)
            .explode("frequencies")
            .assign(
                frequency=lambda df: _expand_df(
                    key="frequency", iterable=df.frequencies
                )
            )
            .assign(
                exchange_asset=lambda df: _expand_df(
                    key="exchange-assets", iterable=df.frequencies
                )
            )
            .explode("exchange_asset")
            .drop("frequencies", axis=1)
            .reset_index(drop=True)
        )
        return convert_catalog_dtypes(df_catalog_metrics)


class CatalogPairMetricsData(List[Any]):
    """
    Transforms catalog pair asset data in list form into a dataframe
    :return: Catalog Data
    """

    def to_dataframe(self) -> DataFrameType:
        df_catalog_metrics = (
            pd.DataFrame(self)
            .explode("frequencies")
            .assign(
                frequency=lambda df: _expand_df(
                    key="frequency", iterable=df.frequencies
                )
            )
            .assign(pair=lambda df: _expand_df(key="pairs", iterable=df.frequencies))
            .explode("pair")
            .drop("frequencies", axis=1)
            .reset_index(drop=True)
        )
        return convert_catalog_dtypes(df_catalog_metrics)


class CatalogInstitutionMetricsData(List[Any]):
    """
    Transforms catalog institution asset data in list form into a dataframe
    :return: Catalog Data
    """

    def to_dataframe(self) -> DataFrameType:
        df_catalog_metrics = (
            pd.DataFrame(self)
            .explode("frequencies")
            .assign(
                frequency=lambda df: _expand_df(
                    key="frequency", iterable=df.frequencies
                )
            )
            .assign(
                institution=lambda df: _expand_df(
                    key="institutions", iterable=df.frequencies
                )
            )
            .explode("institution")
            .drop("frequencies", axis=1)
            .reset_index(drop=True)
        )
        return convert_catalog_dtypes(df_catalog_metrics)


class CatalogMarketMetricsData(List[Any]):
    """
    Transforms catalog data in list form into a dataframe
    :return: Catalog Data
    """

    def to_dataframe(self) -> DataFrameType:
        df_catalog_market_metrics = pd.DataFrame(self)
        df_catalog_market_metrics = (
            df_catalog_market_metrics.explode("metrics")
            .assign(metric=lambda df: _expand_df(key="metric", iterable=df.metrics))
            .assign(
                frequencies=lambda df: _expand_df(
                    key="frequencies", iterable=df.metrics
                )
            )
            .explode("frequencies")
            .assign(
                frequency=lambda df: _expand_df(
                    key="frequency", iterable=df.frequencies
                )
            )
            .assign(
                min_time=lambda df: _expand_df(key="min_time", iterable=df.frequencies)
            )
            .assign(
                max_time=lambda df: _expand_df(key="max_time", iterable=df.frequencies)
            )
            .reset_index(drop=True)
            .drop(["metrics", "frequencies"], axis=1)
        )
        return convert_catalog_dtypes(df_catalog_market_metrics)


class CatalogMarketCandlesData(List[Any]):
    """
    Transforms catalog data in list form into a dataframe
    :return: Catalog Data
    """

    def to_dataframe(self) -> DataFrameType:
        df_catalog_market_candles = pd.DataFrame(self)
        df_catalog_market_candles = (
            df_catalog_market_candles.explode("frequencies")
            .assign(
                frequency=lambda df: _expand_df(
                    key="frequency", iterable=df.frequencies
                )
            )
            .assign(
                min_time=lambda df: _expand_df(key="min_time", iterable=df.frequencies)
            )
            .assign(
                max_time=lambda df: _expand_df(key="max_time", iterable=df.frequencies)
            )
            .reset_index(drop=True)
            .drop(["frequencies"], axis=1)
        )
        return convert_catalog_dtypes(df_catalog_market_candles)


class CatalogAssetPairCandlesData(List[Any]):
    """
    Transforms catalog data in list form into a dataframe
    :return: Catalog Data
    """

    def to_dataframe(self) -> DataFrameType:
        df_catalog_asset_candles = pd.DataFrame(self)
        df_catalog_asset_candles = (
            df_catalog_asset_candles.explode("frequencies")
            .assign(
                frequency=lambda df: _expand_df(
                    key="frequency", iterable=df.frequencies
                )
            )
            .assign(
                min_time=lambda df: _expand_df(key="min_time", iterable=df.frequencies)
            )
            .assign(
                max_time=lambda df: _expand_df(key="max_time", iterable=df.frequencies)
            )
            .reset_index(drop=True)
            .drop(["frequencies"], axis=1)
        )
        return convert_catalog_dtypes(df_catalog_asset_candles)



from copy import deepcopy
from gzip import GzipFile
from io import BytesIO
from logging import getLogger
from time import sleep
from typing import Any, Dict, Iterable, Iterator, List, Optional, cast
from dateutil.parser import isoparse

import requests

from coinmetrics._typing import (
    DataRetrievalFuncType,
    DataReturnType,
    FilePathOrBuffer,
    UrlParamTypes,
    DataFrameType,
)
from coinmetrics._utils import get_file_path_or_buffer

try:
    import orjson as json
except ImportError:
    import json  # type: ignore


logger = getLogger("cm_client_data_collection")

try:
    import pandas as pd  # type: ignore
except ImportError:
    pd = None
    logger.info(
        "Pandas export is unavailable. Install pandas to unlock dataframe functions."
    )


class CsvExportError(Exception):
    pass


class DataFetchError(Exception):
    pass


NUMBER_OF_RETRIES = 3


class DataCollection:
    def __init__(
        self,
        data_retrieval_function: DataRetrievalFuncType,
        endpoint: str,
        url_params: Dict[str, UrlParamTypes],
        csv_export_supported: bool = True,
    ) -> None:
        self._csv_export_supported = csv_export_supported
        self._data_retrieval_function = data_retrieval_function
        self._endpoint = endpoint
        self._url_params = url_params
        self._next_page_token: Optional[str] = ""
        self._current_data_iterator: Optional[Iterator[Any]] = None

    def first_page(self) -> List[Dict[str, Any]]:
        return cast(
            List[Dict[str, Any]],
            self._fetch_data_with_retries(self._url_params)["data"],
        )

    def to_list(self) -> List[Dict[str, Any]]:
        return [data for data in self]

    def __next__(self) -> Any:
        try:
            if self._current_data_iterator is not None:
                return next(self._current_data_iterator)
        except StopIteration:
            if self._next_page_token is None:
                raise StopIteration

        url_params = deepcopy(self._url_params)
        if self._next_page_token:
            url_params["next_page_token"] = self._next_page_token
        api_response = self._data_retrieval_function(self._endpoint, url_params)
        self._next_page_token = cast(Optional[str], api_response.get("next_page_token"))
        self._current_data_iterator = iter(api_response["data"])
        return next(self._current_data_iterator)

    def __iter__(self) -> "DataCollection":
        return self

    def _fetch_data_with_retries(
        self, url_params: Dict[str, UrlParamTypes]
    ) -> DataReturnType:
        for i in range(1, NUMBER_OF_RETRIES + 1):
            try:
                return self._data_retrieval_function(self._endpoint, url_params)
            except requests.exceptions.ProxyError as e:
                if i == NUMBER_OF_RETRIES:
                    raise
                logger.warning(
                    "failed to fetch data with error: %s, retrying in 1 second, try (%s/%s)",
                    e,
                    i,
                    NUMBER_OF_RETRIES,
                )
                sleep(1)

        raise DataFetchError("Failed to fetch data")

    def export_to_csv(
        self,
        path_or_bufstr: FilePathOrBuffer = None,
        columns_to_store: Optional[List[str]] = None,
        compress: bool = False,
    ) -> Optional[str]:
        if not self._csv_export_supported:
            raise CsvExportError(
                "Sorry, csv export is not supported for this data type."
            )

        return self._export_to_file(
            self._get_csv_data_lines(columns_to_store), path_or_bufstr, compress
        )

    def _get_csv_data_lines(
        self, columns_to_store: Optional[List[str]]
    ) -> Iterable[bytes]:
        first_data_el = None
        if columns_to_store is None:
            try:
                first_data_el = next(self)
            except StopIteration:
                logger.info("no data to export")
                return
            columns_to_store = list(first_data_el.keys())

        yield (",".join(columns_to_store) + "\n").encode()

        if first_data_el is not None:
            yield (
                ",".join(
                    f'"{first_data_el.get(column)}"' or ""
                    for column in columns_to_store
                )
                + "\n"
            ).encode()
        for data_el in self:
            yield (
                ",".join(
                    f'"{data_el.get(column)}"' or "" for column in columns_to_store
                )
                + "\n"
            ).encode()

    def export_to_json(
        self,
        path_or_bufstr: FilePathOrBuffer = None,
        compress: bool = False,
    ) -> Optional[str]:
        def _gen_json_lines() -> Iterable[bytes]:
            for data_row in self:
                yield json.dumps(data_row) + b"\n"

        return self._export_to_file(_gen_json_lines(), path_or_bufstr, compress)

    def _export_to_file(
        self,
        data_generator: Iterable[bytes],
        path_or_bufstr: FilePathOrBuffer = None,
        compress: bool = False,
    ) -> Optional[str]:

        if path_or_bufstr is None:
            path_or_bufstr_obj: FilePathOrBuffer = BytesIO()
        else:
            path_or_bufstr_obj = path_or_bufstr

        path_or_bufstr_obj = get_file_path_or_buffer(path_or_bufstr_obj)
        if hasattr(path_or_bufstr_obj, "write"):
            f = path_or_bufstr_obj
            close = False
        else:
            f = open(path_or_bufstr_obj, "wb")  # type: ignore
            close = True
        if compress:
            output_file = GzipFile(fileobj=f)  # type: ignore
        else:
            output_file = f  # type: ignore
        try:
            for line in data_generator:
                output_file.write(line)
        finally:
            if compress:
                output_file.close()
            if close:
                f.close()  # type: ignore

        if path_or_bufstr is None:
            return path_or_bufstr_obj.getvalue().decode()  # type: ignore

        return None

    def to_dataframe(
        self,
        header: Optional[List[str]] = None,
        dtype_mapper: Optional[Dict[str, Any]] = None,
        optimize_pandas_types: Optional[bool] = True,
    ) -> DataFrameType:
        """
        Outputs a pandas dataframe.

        :param header: Optional column names for outputted dataframe. List length must match the output.
        :type header: list(str)
        :param dtype_mapper: Dictionary for converting columns to types, where keys are columns and values are types that pandas accepts in an `as_type()` call. This mapping is prioritized over the pandas dtype conversions.
        :type dtype_mapper: dict
        :param optimize_pandas_types: Boolean flag for using pandas typing for dataframes. If True and dtype_mapper is not specified, then the output will result in all columns with string dtype. If True and dtype_mapper is specified, then the output will convert the dataframe columns with the user-specified dtyoes.
        :type optimize_pandas_types: Bool
        :return: Data in a pandas dataframe
        :rtype: DataFrameType
        """
        if pd is None:
            logger.info("Pandas not found; Returning None")
            return None
        else:
            if optimize_pandas_types:
                f = BytesIO()
                self.export_to_csv(f)
                if f.getbuffer().nbytes == 0:
                    logger.warning("Response is empty.")
                    return pd.DataFrame()
                else:
                    f.seek(0)
                    columns = (
                        BytesIO(f.getvalue())
                        .readlines(1)[0]
                        .decode()
                        .strip()
                        .split(",")
                    )
                    datetime_cols = [
                        c for c in columns if c.endswith("_time") or c == "time"
                    ]
                    df = pd.read_csv(
                        f,
                        parse_dates=datetime_cols,
                        dtype=dtype_mapper,
                        date_parser=isoparse,
                    )
                    if dtype_mapper is None:
                        df = df.convert_dtypes()
                    if header is not None:
                        assert len(df.columns) == len(
                            header
                        ), "header length does not match output values"
                        df.columns = header
                    return df
            else:
                if dtype_mapper is None:
                    return pd.DataFrame(self)
                else:
                    return pd.DataFrame(self).astype(dtype=dtype_mapper)



import pathlib
from datetime import date, datetime
from enum import Enum
from functools import wraps
from logging import getLogger
from os.path import expanduser
from time import sleep
from typing import Any, Callable, Dict, List, Optional, Tuple, Union
from coinmetrics._typing import FilePathOrBuffer, UrlParamTypes

logger = getLogger("cm_client_utils")


def transform_url_params_values_to_str(
    params: Dict[str, UrlParamTypes]
) -> Dict[str, str]:
    processed_params = {}
    for param_name, param_value in params.items():
        if param_value is None:
            continue
        if isinstance(param_value, (datetime, date)):
            if param_name.endswith("_time"):
                processed_params[param_name] = param_value.isoformat()
            else:
                raise ValueError(
                    "`{}` doesn't support {} objects".format(
                        param_name, type(param_value)
                    )
                )
        elif isinstance(param_value, (list, tuple)):
            processed_params[param_name] = ",".join(param_value)
        elif isinstance(param_value, Enum):
            processed_params[param_name] = param_value.value
        elif isinstance(param_value, bool):
            processed_params[param_name] = "true" if param_value else "false"
        else:
            processed_params[param_name] = str(param_value)
    return processed_params


def get_file_path_or_buffer(filepath_or_buffer: FilePathOrBuffer) -> FilePathOrBuffer:
    if isinstance(filepath_or_buffer, (str, bytes, pathlib.Path)):
        return _stringify_path(filepath_or_buffer)

    if not _is_file_like(filepath_or_buffer):
        msg = "Invalid file path or buffer object type: {}".format(
            type(filepath_or_buffer)
        )
        raise ValueError(msg)

    return filepath_or_buffer


def _stringify_path(filepath_or_buffer: FilePathOrBuffer) -> FilePathOrBuffer:
    if hasattr(filepath_or_buffer, "__fspath__"):
        # https://github.com/python/mypy/issues/1424
        return filepath_or_buffer.__fspath__()  # type: ignore
    elif isinstance(filepath_or_buffer, pathlib.Path):
        return str(filepath_or_buffer)
    return _expand_user(filepath_or_buffer)


def _expand_user(
    filepath_or_buffer: FilePathOrBuffer,
) -> FilePathOrBuffer:
    if isinstance(filepath_or_buffer, str):
        return expanduser(filepath_or_buffer)
    return filepath_or_buffer


def _is_file_like(obj: Any) -> bool:
    if not (hasattr(obj, "read") or hasattr(obj, "write")):
        return False

    if not hasattr(obj, "__iter__"):
        return False

    return True


def retry(
    error_cls: Union[List[Any], Tuple[Any], Any],
    retries: int = 5,
    wait_time_between_retries: int = 30,
    message: Optional[str] = None,
    fail: bool = True,
    error_str: Optional[str] = None,
) -> Callable[..., Any]:
    def retry_wrapper(f: Any) -> Any:
        @wraps(f)
        def wrapper(*args: List[Any], **kwargs: Dict[str, Any]) -> Any:
            for n in range(1, retries + 1):
                try:
                    return f(*args, **kwargs)
                except error_cls as error:  # type: ignore
                    if (
                        n == retries
                        or (error_str is not None and error_str not in str(error))
                    ) and fail:
                        raise
                    if callable(wait_time_between_retries):
                        wait_time = wait_time_between_retries()
                    else:
                        wait_time = wait_time_between_retries
                    if message:
                        logger.info(
                            "%s. Retrying in: %s sec. Iteration: %s",
                            message,
                            wait_time,
                            n,
                        )

                    sleep(wait_time)

        return wrapper

    return retry_wrapper



import pandas as pd  # type: ignore
import pytest
from coinmetrics.api_client import CoinMetricsClient
import os

CM_API_KEY = os.environ.get("CM_API_KEY")
client = CoinMetricsClient(str(CM_API_KEY))
cm_api_key_set = CM_API_KEY is not None
REASON_TO_SKIP = "Need to set CM_API_KEY as an env var in order to run this test"


@pytest.mark.skipif(not cm_api_key_set, reason=REASON_TO_SKIP)
def test_secondary_level_catalog_assets() -> None:
    """
    Tests that the secondary expansion functions works as expected - there is one row created for each secondary level
    expansion
    """
    catalog_assets: pd.DataFrame = client.catalog_assets().to_dataframe()
    catalog_assets_metrics_expansion: pd.DataFrame = (
        client.catalog_assets().to_dataframe(secondary_level="metrics")
    )
    catalog_assets_usdc_metrics = catalog_assets[
        catalog_assets["asset"] == "usdc"
    ].iloc[0]["metrics"]
    usdc_metrics = catalog_assets_metrics_expansion[
        catalog_assets_metrics_expansion["asset"] == "usdc"
    ]["metric"]
    usdc_metrics_unique = set(usdc_metrics)
    assert len(usdc_metrics_unique) == len(catalog_assets_usdc_metrics)
    assert len(usdc_metrics_unique) > 100

    catalog_assets_markets_expansion: pd.DataFrame = (
        client.catalog_assets().to_dataframe(secondary_level="markets")
    )
    catalog_assets_usdc_markets = catalog_assets[
        catalog_assets["asset"] == "usdc"
    ].iloc[0]["markets"]
    usdc_markets = catalog_assets_markets_expansion[
        catalog_assets_markets_expansion["asset"] == "usdc"
    ]["market"]
    usdc_markets_unqiue = set(usdc_markets)
    assert len(usdc_markets_unqiue) == len(catalog_assets_usdc_markets)


@pytest.mark.skipif(not cm_api_key_set, reason=REASON_TO_SKIP)
def test_secondary_level_catalog_morderbooks() -> None:
    """
    Tests that the binance-BTCUSDT-future expands into multiple rows as expected
    """
    market_to_test = "binance-BTCUSDT-future"
    catalog_orderbooks = client.catalog_market_orderbooks().to_dataframe()
    catalog_orderbooks_depths_expansion = (
        client.catalog_market_orderbooks().to_dataframe(secondary_level="depths")
    )
    catalog_orderbooks_market_depths = catalog_orderbooks[
        catalog_orderbooks["market"] == market_to_test
    ].iloc[0]["depths"]
    catalog_orderbooks_depths_expansion_market = catalog_orderbooks_depths_expansion[
        catalog_orderbooks_depths_expansion["market"] == market_to_test
    ]
    assert len(catalog_orderbooks_market_depths) == len(
        catalog_orderbooks_depths_expansion_market
    )
    assert len(catalog_orderbooks_market_depths) > 1


@pytest.mark.skipif(not cm_api_key_set, reason=REASON_TO_SKIP)
def test_secondary_level_catalog_asset_alerts() -> None:
    """
    Probably not worth implementing secondary level in this case - this test just confirms this doesn't break, it
    previously was broken due to changes in data model
    """
    catalog_asset_alerts = client.catalog_asset_alerts().to_dataframe()
    conditions = list(catalog_asset_alerts["conditions"])
    for item in conditions:
        print(item)
    print("test!")


@pytest.mark.skipif(not cm_api_key_set, reason=REASON_TO_SKIP)
def test_catalog_asset_pairs_to_df() -> None:
    """
    Just a sanity test, no relevant secondary level here
    """
    catalog_asset_pairs: pd.DataFrame = client.catalog_asset_pairs().to_dataframe()
    assert len(catalog_asset_pairs) > 5


@pytest.mark.skipif(not cm_api_key_set, reason=REASON_TO_SKIP)
def test_catalog_exchanges_secondary_level() -> None:
    """
    Checks that the binance exchange is expanded the expected way - for metrics same amount of rows as there are unique
    metric and same amount of rows as there are markets
    """
    exchange_to_test = "binance"
    catalog_exchanges = client.catalog_exchanges().to_dataframe()
    catalog_exchanges_metrics_expansion = client.catalog_exchanges().to_dataframe(
        secondary_level="metrics"
    )
    catalog_exchanges_metrics = catalog_exchanges[
        catalog_exchanges["exchange"] == exchange_to_test
    ].iloc[0]["metrics"]
    metrics_expansion_metrics = set(
        catalog_exchanges_metrics_expansion[
            catalog_exchanges_metrics_expansion["exchange"] == exchange_to_test
        ]["metric"]
    )
    assert len(catalog_exchanges_metrics) == len(metrics_expansion_metrics)
    assert len(metrics_expansion_metrics) > 1

    catalog_exchange_markets = catalog_exchanges[
        catalog_exchanges["exchange"] == exchange_to_test
    ].iloc[0]["markets"]
    catalog_exchange_markets_expansion = client.catalog_exchanges().to_dataframe(
        secondary_level="markets"
    )
    markets_expansion_markets = catalog_exchange_markets_expansion[
        catalog_exchange_markets_expansion["exchange"] == exchange_to_test
    ]["market"]
    assert len(catalog_exchange_markets) == len(markets_expansion_markets)


@pytest.mark.skipif(not cm_api_key_set, reason=REASON_TO_SKIP)
def test_catalog_asset_metrics() -> None:
    """
    No extensive testing needed, just checks that a dataframe is created
    """
    asset_metrics = client.catalog_asset_metrics().to_dataframe()
    assert len(asset_metrics) > 100


@pytest.mark.skipif(not cm_api_key_set, reason=REASON_TO_SKIP)
def test_catalog_exchange_metrics() -> None:
    """
    No extensive testing is needed, just checks dataframe is created
    """
    exchange_metrics = client.catalog_exchange_metrics().to_dataframe()
    assert len(exchange_metrics) > 10


@pytest.mark.skipif(not cm_api_key_set, reason=REASON_TO_SKIP)
def test_catalog_exchange_asset_metrics() -> None:
    """
    No extensive testing needed, just checks dataframe is created
    """
    exchange_asset_metrics = client.catalog_exchange_asset_metrics().to_dataframe()
    assert len(exchange_asset_metrics) > 20


@pytest.mark.skipif(not cm_api_key_set, reason=REASON_TO_SKIP)
def test_catalog_pair_metrics() -> None:
    """
    No extensive testing needed, just checks dataframe is created
    """
    pair_metrics = client.catalog_pair_metrics().to_dataframe()
    assert len(pair_metrics) > 100
    print("test!")


@pytest.mark.skipif(not cm_api_key_set, reason=REASON_TO_SKIP)
def test_catalog_institution_metrics() -> None:
    """
    No extensive testing needed, just checks dataframe is created
    """
    inst_metrics = client.catalog_institution_metrics().to_dataframe()
    assert len(inst_metrics) > 20


@pytest.mark.skipif(not cm_api_key_set, reason=REASON_TO_SKIP)
def test_catalog_asset_pair_candles() -> None:
    """
    No extensive testing needed, just checks dataframe is created
    """
    pair_candles = client.catalog_asset_pair_candles().to_dataframe()
    assert len(pair_candles) > 10


@pytest.mark.skipif(not cm_api_key_set, reason=REASON_TO_SKIP)
def test_catalog_indexes() -> None:
    """
    No extensive testing needed, just checks dataframe is created
    """
    indexes = client.catalog_indexes().to_dataframe()
    assert len(indexes) > 10


@pytest.mark.skipif(not cm_api_key_set, reason=REASON_TO_SKIP)
def test_catalog_index_candles() -> None:
    """
    No extensive testing needed, just checks dataframe is created
    """
    indexes = client.catalog_index_candles().to_dataframe()
    assert len(indexes) > 10


@pytest.mark.skipif(not cm_api_key_set, reason=REASON_TO_SKIP)
def test_catalog_institutions() -> None:
    """
    No extensive testing needed, just checks dataframe is created
    """
    institutions = client.catalog_institutions().to_dataframe()
    assert len(institutions) > 10


@pytest.mark.skipif(not cm_api_key_set, reason=REASON_TO_SKIP)
def test_catalog_markets() -> None:
    """
    No extensive testing needed, just checks dataframe is created
    """
    catalog_markets = client.catalog_markets(exchange="coinbase").to_dataframe()
    assert len(catalog_markets) > 10


@pytest.mark.skipif(not cm_api_key_set, reason=REASON_TO_SKIP)
def test_catalog_market_trades() -> None:
    """
    No extensive testing needed, just checks dataframe is created
    """
    catalog_market_trades = client.catalog_market_trades().to_dataframe()
    assert len(catalog_market_trades) > 20


@pytest.mark.skipif(not cm_api_key_set, reason=REASON_TO_SKIP)
def test_catalog_metrics() -> None:
    """
    No extensive testing needed, just checks dataframe is created
    """
    catalog_metrics = client.catalog_metrics().to_dataframe()
    assert len(catalog_metrics) > 100


@pytest.mark.skipif(not cm_api_key_set, reason=REASON_TO_SKIP)
def test_catalog_market_metrics() -> None:
    """
    No extensive testing needed, just checks dataframe is created
    """
    catalog_market_metrics = client.catalog_market_metrics(
        exchange="coinbase"
    ).to_dataframe()
    assert len(catalog_market_metrics) > 20


@pytest.mark.skipif(not cm_api_key_set, reason=REASON_TO_SKIP)
def test_catalog_market_candles() -> None:
    """
    No extensive testing needed, just checks dataframe is created
    """
    catalog_market_candles = client.catalog_market_candles(
        exchange="coinbase"
    ).to_dataframe()
    assert len(catalog_market_candles) > 50


@pytest.mark.skipif(not cm_api_key_set, reason=REASON_TO_SKIP)
def test_catalog_market_quotes() -> None:
    """
    No extensive testing needed, just checks dataframe is created
    """
    catalog_market_quotes = client.catalog_market_quotes().to_dataframe()
    assert len(catalog_market_quotes) > 100


@pytest.mark.skipif(not cm_api_key_set, reason=REASON_TO_SKIP)
def test_catalog_market_funding_rates() -> None:
    """
    No extensive testing needed, just checks dataframe is created
    """
    catalog_market_funding_rates = client.catalog_market_funding_rates().to_dataframe()
    assert len(catalog_market_funding_rates) > 100


@pytest.mark.skipif(not cm_api_key_set, reason=REASON_TO_SKIP)
def test_catalog_market_greeks() -> None:
    """
    No extensive testing needed, just checks dataframe is created
    """
    catalog_market_greeks = client.catalog_market_greeks().to_dataframe()
    assert len(catalog_market_greeks) > 100


@pytest.mark.skipif(not cm_api_key_set, reason=REASON_TO_SKIP)
def test_catalog_market_openinterest() -> None:
    """
    No extensive testing needed, just checks dataframe is created
    """
    catalog_open_interest = client.catalog_market_open_interest().to_dataframe()
    assert len(catalog_open_interest) > 100


@pytest.mark.skipif(not cm_api_key_set, reason=REASON_TO_SKIP)
def test_catalog_market_liquidations() -> None:
    """
    No extensive testing needed, just checks dataframe is created
    """
    catalog_liquidations = client.catalog_market_liquidations().to_dataframe()
    assert len(catalog_liquidations) > 100


@pytest.mark.skipif(not cm_api_key_set, reason=REASON_TO_SKIP)
def test_defi_balance_sheets() -> None:
    """
    Testing this works for aave
    """
    aave_balance_sheets = client.get_defi_balance_sheets(
        defi_protocols="aave_v2_eth"
    ).to_dataframe()
    assert len(aave_balance_sheets) > 100
    assert aave_balance_sheets.iloc[:1]["defi_protocol"][0] == "aave_v2_eth"


if __name__ == "__main__":
    pytest.main()



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['coinmetrics']

package_data = \
{'': ['*']}

install_requires = \
['orjson>=3.6.0,<4.0.0',
 'python-dateutil>=2.8.2,<3.0.0',
 'requests>=2.24.0,<3.0.0',
 'typer>=0.6.1,<0.7.0',
 'websocket-client>=1.2.1,<2.0.0']

extras_require = \
{'pandas': ['pandas>=1.3.3,<2.0.0']}

entry_points = \
{'console_scripts': ['coinmetrics = coinmetrics.typer_cli:main']}

setup_kwargs = {
    'name': 'coinmetrics-api-client',
    'version': '2023.2.27.22',
    'description': 'Python client for Coin Metrics API v4.',
    'long_description': '# Coin Metrics Python API v4 client library\n\nThis is an official Python API client for Coin Metrics API v4.\n\n## Installation and Updates\nTo install the client you can run the following command:\n```\npip install coinmetrics-api-client\n```\n\nNote that the client is updated regularly to reflect the changes made in [API v4](https://docs.coinmetrics.io/api/v4). Ensure that your latest version matches with what\'s in [pyPI](https://pypi.org/project/coinmetrics-api-client/) \n\nTo update your version, run the following command:\n```\npip install coinmetrics-api-client -U\n```\n\n## Introduction\nYou can use this client for querying all kinds of data with your API.\n\nTo initialize the client you should use your API key, and the CoinMetricsClient class like the following.\n```\nfrom coinmetrics.api_client import CoinMetricsClient\n\nclient = CoinMetricsClient("<cm_api_key>")\n\n# or to use community API:\nclient = CoinMetricsClient()\n```\n\nAfter that you can use the client object for getting information such as available markets:\n```\nprint(client.catalog_markets())\n```\n\nor to query all available assets along with what is available for those assets, like metrics, markets:\n\n```\nprint(client.catalog_assets())\n```\n\n\nyou can also use filters for the catalog endpoints like this:\n\n```\nprint(client.catalog_assets(assets=[\'btc\']))\n```\nin this case you would get all the information for btc only. \n\nYou can use this client to connect to our API v4 and get catalog or timeseries data from python environment. It natively supports paging over the data so you can use it to iterate over timeseries entries seamlessly.\n\nThe client can be used to query both pro and community data.\n\nThe full list of methods can be found in the [API Client Spec](https://coinmetrics.github.io/api-client-python/site/api_client.html).\n\n\nIf you\'d like a more wholistic view of what is offered from an API endpoint you can use the `to_dataframe()` function \nassociated with our catalog endpoints. The code snippet below shows getting a dataframe of information on all the \nassets that data is provided for:\n```python\nprint(client.catalog_assets().to_dataframe())\n```\n\nOutput:\n```commandline\n      asset          full_name          exchanges  ... metrics atlas  experimental\n0      100x           100xCoin          [gate.io]  ...     NaN  <NA>          <NA>\n1     10set             Tenset   [gate.io, lbank]  ...     NaN  <NA>          <NA>\n2       18c           Block 18            [huobi]  ...     NaN  <NA>          <NA>\n3      1art          ArtWallet          [gate.io]  ...     NaN  <NA>          <NA>\n4      1box               1BOX           [zb.com]  ...     NaN  <NA>          <NA>\n```\n\nNow you can use the pandas Dataframe functionality to do useful transformations, such as filtering out the assets \nwithout metrics available, then saving that data to a csv file:\n```python\ncatalog_assets_df = client.catalog_assets().to_dataframe()\nonly_assets_with_metrics = catalog_assets_df.dropna(subset=[\'metrics\'])\nonly_assets_with_metrics.to_csv("cm_assets_with_metrics.csv")\n```\n\nYou may notice that in that data saved, the "metrics" column for example is a list of json data describing the metrics \noffered and the frequency at which they are available. To help parse this information there is a keyword for all catalog\nendpoint data `secondary_level`:\n```python\ncatalog_assets_df = client.catalog_assets().to_dataframe(secondary_level="metrics")\nonly_assets_with_metrics = catalog_assets_df.dropna(subset=[\'metric\'])\neth_metrics = only_assets_with_metrics[only_assets_with_metrics[\'asset\'] == "eth"]\neth_metrics.to_csv("eth_metrics_granular.csv")\n```\n\nThe above example queries for eth metrics at the level of metrics and frequency, where it will have one row for each \nmetric and frequency related to Ethereum. This allows users to quickly get high level information about exactly what\nis offered from the Coin Metrics API and to make custom queries against the API from there. This example only covers \n`catalog_assets()`, but the pattern can be used across all of our catalog endpoints. \n\n\n## Examples\nThe API Client allows you to chain together workflows for importing, transforming, then exporting Coin Metrics data.\nBelow are examples of common use-cases that can be altered to tailor your specific needs. In addition to the examples \nlisted below, there\'s examples covering all the API methods, found in the [examples directory](https://github.com/coinmetrics/api-client-python/tree/master/examples).\n\n**[Example Notebooks](https://github.com/coinmetrics/api-client-python/tree/master/examples/notebooks)**\n\n* `walkthrough_community.ipynb`: Walks through the basic functionality available using the community client.\n\n**[Asset Metrics](https://github.com/coinmetrics/api-client-python/tree/master/examples/asset_metrics)**\n\n* `bbb_metrics_csv_exporter_using_plain_requests.py`: Queries block-by-block metrics using the `requests` library and exports the output into a CSV file.\n* `bbb_metrics_json_exporter.py`: Queries block-by-block metrics and exports the output into a JSON file.\n* `eod_metrics_csv_exporter.py`: Exports a set of user-defined metrics and assets published at end-of-day and exports the output into a CSV file.\n* `reference_rates_json_exporter.py`: Queries Coin Metrics Reference Rates at a user-defined frequency for a set of assets, then exports the output into a JSON file.\n\n**[Market Data](https://github.com/coinmetrics/api-client-python/tree/master/examples/market_data)** \n\n* `books_json_exporter.py`: Queries market orderbook data then exports the output into a JSON file.\n* `candles_json_exporter.py`: Queries market candles data then exports the output into a JSON file.\n* `funding_rates_json_exporter.py`: Queries market funding rates data then exports the output into a JSON file.\n* `trades_csv_exporter.py`: Queries market trades data then exports the output into a CSV file.\n* `trades_json_exporter.py`: Queries market trades data then exports the output into a JSON file.\n\n## Getting timeseries data\n\nFor getting timeseries data you want to use methods of the client class that start with `get_`. It\'s important to note\nthat the timeseries endpoints return data of a parent class type `DataCollection`. The `DataCollection` class is meant\nto support a variety of different data export and data manipulation use cases, so just calling one of the client\nmethods such as `data = client.get_market_trades(markets="coinbase-btc-usd-spot")` will not actually retrieve the data related\nto this API call. You must then call a function on this `DataCollection` such as `data.export_to_csv("coinbase_btc_usd_spot_trades.csv)`\nor `data.to_dataframe()` in order to access the data. There is more explicit examples below.If you are curious to see\nhow the API calls are being made and with what parameters, instantiating the client with the `verbose` argument like \n`CoinMetricsClient(api_key=<YOUR_API_KEY>, verbose=True)` will print the API calls as well as information on performance to console. \n\nFor example if you want to get a bunch of market data trades for coinbase btc-usd pair you can run something similar to the following:\n\n```\nfor trade in client.get_market_trades(\n    markets=\'coinbase-btc-usd-spot\', \n    start_time=\'2020-01-01\', \n    end_time=\'2020-01-03\',\n    limit_per_market=10\n):\n    print(trade)\n```\nThis example uses the `DataCollection` as a Python iterator, so with each iteration of the Python for loop it will\ncall the Coin Metrics API and return data. The default `page_size` for calls to the API is 100, so each call will return\n100 trades until it reaches the end of the query. To get more trades in each API call, you can add the parameter\n`page_size` to the `.get_market_trades()` method call, up to a maximum of 10000. \n\nOr if you want to see daily btc asset metrics you can use something like this:\n\n```\nfor metric_data in client.get_asset_metrics(assets=\'btc\', \n                                            metrics=[\'ReferenceRateUSD\', \'BlkHgt\', \'AdrActCnt\',  \n                                                     \'AdrActRecCnt\', \'FlowOutBFXUSD\'], \n                                            frequency=\'1d\',\n                                            limit_per_asset=10):\n    print(metric_data)\n```\nThis will print you the requested metrics for all the days where we have any of the metrics present.\n\n\n### DataFrames\n_(New in >=`2021.9.30.14.30`)_\n\nTimeseries data can be transformed into a pandas dataframe by using the `to_dataframe()` method. The code snippet below shows how:\n```\nimport pandas as pd\nfrom coinmetrics.api_client import CoinMetricsClient\nfrom os import environ\n\nclient = CoinMetricsClient()\ntrades = client.get_market_trades(\n    markets=\'coinbase-btc-usd-spot\', \n    start_time=\'2021-09-19T00:00:00Z\', \n    limit_per_market=10\n)\ntrades_df = trades.to_dataframe()\nprint(trades_df.head())\n\n```\nIf you want to use dataframes, then you will need to install pandas\n\n**Notes**\n\n- This only works with requests that return the type `DataCollection`. Thus, `catalog` requests, which return lists cannot be returned as dataframes.\n  Please see the [API Client Spec](https://coinmetrics.github.io/api-client-python/site/api_client.html) for a full list\n  of requests and their return types.\n- API restrictions apply. Some requests may return empty results due to limited access to the API from you API key.\n\n#### Type Conversion \n_(New in >=`2021.12.17.18.00`)_\n\nAs of version `2021.12.17.18.00` or later, outputs from the  `to_dataframe` function automatically convert the dtypes for a dataframe to the optimal pandas types.\n```python\nmetrics_list = [\'volume_trusted_spot_usd_1d\', \'SplyFF\', \'AdrBalUSD1Cnt\']\nasset_list = [\'btc\',\'xmr\']\nstart_time=\'2021-12-01\'\ndf_metrics = client.get_asset_metrics(\n  assets=asset_list, metrics=metrics_list, start_time=start_time, limit_per_asset=3\n).to_dataframe()\nprint(df_metrics.dtypes)\n```\n\n```\nasset                          string\ntime                           datetime64[ns, tzutc()]\nAdrBalUSD1Cnt                   Int64\nSplyFF                        Float64\nvolume_trusted_spot_usd_1d    Float64\ndtype: object\n```\n\nThis can be turned off by setting `optimize_pandas_types=False`\n\nAlternatively, you can manually enter your own type conversion by passing in a dictionary for `dtype_mapper`. This can be done in conjunction with pandas\' built in type optimizations.\n```python\nmapper = {\n  \'SplyFF\': \'Float64\',\n  \'AdrBalUSD1Cnt\': \'Int64\',\n}\ndf_mapped = client.get_asset_metrics(\n  assets=asset_list, metrics=metrics_list, start_time=start_time, limit_per_asset=3\n).to_dataframe(dtype_mapper=mapper, optimize_pandas_types=True)\nprint(df_mapped.dtypes)\n```\n\n```\nasset                                          object\ntime                          datetime64[ns, tzutc()]\nAdrBalUSD1Cnt                                   Int64\nSplyFF                                        Float64\nvolume_trusted_spot_usd_1d                    float64\ndtype: object\n```\n\nOr as strictly the only types in the dataframe\n\n```python\ndtype_mapper = {\n    \'ReferenceRateUSD\': np.float64,\n    \'time\': np.datetime64\n}\ndf = client.get_asset_metrics(\n  assets=\'btc\', metrics=\'ReferenceRateUSD\', start_time=\'2022-06-15\', limit_per_asset=1\n).to_dataframe(dtype_mapper=dtype_mapper, optimize_pandas_types=False)\ndf.info()\n```\n```\nRangeIndex: 1 entries, 0 to 0\nData columns (total 3 columns):\n #   Column            Non-Null Count  Dtype         \n---  ------            --------------  -----         \n 0   asset             1 non-null      object        \n 1   time              1 non-null      datetime64[ns]\n 2   ReferenceRateUSD  1 non-null      float64       \ndtypes: datetime64[ns](1), float64(1), object(1)\nmemory usage: 152.0+ bytes\n```\n\nNote that in order to pass a custom datetime object, setting a dtype_mapper is mandatory.\n\nPandas type conversion tends to be more performant. But if there are custom operations that must be done using numpy datatypes, this option will let you perform them.\n\n### Exporting to csv and json files:\nYou can also easily export timeseries data to csv and json files with builtin functions on the `DataCollection` type. \nFor example this script will export Coinbase btc and eth trades for a date to csv and json files respectively:\n```python\n    start_date = datetime.date(year=2022, month=1, day=1)\n    end_date = datetime.datetime(year=2022, month=1, day=1)\n    market_trades_btc = client.get_market_trades(page_size=1000, markets="coinbase-btc-usd-spot", start_time=start_date, end_time=end_date)\n    market_trades_btc.export_to_csv("jan_1_2022_coinbase_btc_trades.csv")\n    market_trades_eth = client.get_market_trades(page_size=1000, markets="coinbase-eth-usd-spot", start_time=start_date, end_time=end_date)\n    market_trades_eth.export_to_json("jan_1_2022_coinbase_eth.json")\n```\n\n### Paging\nYou can make the datapoints to iterate from start (default) or from end.\n\nfor that you should use a paging_from argument like the following:\n```\nfrom coinmetrics.api_client import CoinMetricsClient\nfrom coinmetrics.constants import PagingFrom\n\nclient = CoinMetricsClient()\n\nfor metric_data in client.get_asset_metrics(assets=\'btc\', metrics=[\'ReferenceRateUSD\'],\n                                            paging_from=PagingFrom.START):\n    print(metric_data)\n```\n\nPagingFrom.END: is available but by default it will page from the start.\n\n\n### Debugging the API Client\nThere are two additional options for the API Client - `debug_mode` and `verbose`. These two options log network calls \nto the console, and in the case of `debug_mode` it will generate a log file of all the network requests and the time\nit takes to call them. These tools can be used to diagnose issues in your code and also to get a better understanding \nof request times so that users can write more performant code. For example, running the below code:\n```python\nimport os\n\nfrom coinmetrics.api_client import CoinMetricsClient\n\napi_key = os.environ[\'CM_API_KEY\']\n\nif __name__ == \'__main__\':\n    client = CoinMetricsClient(api_key=api_key, debug_mode=True)\n    reference_rates_example = client.get_asset_metrics(assets=[\'btc\', \'algo\', \'eth\'], metrics=[\'ReferenceRateUSD\'])\n    for data in reference_rates_example:\n        continue\n```\n\nThe console output will look like:\n```commandline\n[DEBUG] 2023-01-09 11:01:02,044 - Starting API Client debugging session. logging to stdout and cm_api_client_debug_2023_01_09_11_01_02.txt\n[DEBUG] 2023-01-09 11:01:02,044 - Using coinmetrics version 2022.11.14.16\n[DEBUG] 2023-01-09 11:01:02,044 - Current state of API Client, excluding API KEY: {\'_verify_ssl_certs\': True, \'_api_base_url\': \'https://api.coinmetrics.io/v4\', \'_ws_api_base_url\': \'wss://api.coinmetrics.io/v4\', \'_http_header\': {\'Api-Client-Version\': \'2022.11.14.16\'}, \'_proxies\': {\'http\': None, \'https\': None}, \'debug_mode\': True, \'verbose\': False}\n[DEBUG] 2023-01-09 11:01:02,044 - Attempting to call url: timeseries/asset-metrics with params: {\'assets\': [\'btc\', \'algo\', \'eth\'], \'metrics\': [\'ReferenceRateUSD\'], \'frequency\': None, \'page_size\': None, \'paging_from\': \'start\', \'start_time\': None, \'end_time\': None, \'start_height\': None, \'end_height\': None, \'start_inclusive\': None, \'end_inclusive\': None, \'timezone\': None, \'sort\': None, \'limit_per_asset\': None}\n[DEBUG] 2023-01-09 11:01:02,387 - Response status code: 200 for url: https://api.coinmetrics.io/v4/timeseries/asset-metrics?api_key=[REDACTED]&assets=btc%2Calgo%2Ceth&metrics=ReferenceRateUSD&paging_from=start took: 0:00:00.342874 response body size (bytes): 9832\n[DEBUG] 2023-01-09 11:01:02,388 - Attempting to call url: timeseries/asset-metrics with params: {\'assets\': [\'btc\', \'algo\', \'eth\'], \'metrics\': [\'ReferenceRateUSD\'], \'frequency\': None, \'page_size\': None, \'paging_from\': \'start\', \'start_time\': None, \'end_time\': None, \'start_height\': None, \'end_height\': None, \'start_inclusive\': None, \'end_inclusive\': None, \'timezone\': None, \'sort\': None, \'limit_per_asset\': None, \'next_page_token\': \'0.MjAxOS0wOS0zMFQwMDowMDowMFo\'}\n[DEBUG] 2023-01-09 11:01:02,559 - Response status code: 200 for url: https://api.coinmetrics.io/v4/timeseries/asset-metrics?api_key=[REDACTED]&assets=btc%2Calgo%2Ceth&metrics=ReferenceRateUSD&paging_from=start&next_page_token=0.MjAxOS0wOS0zMFQwMDowMDowMFo took: 0:00:00.171487 response body size (bytes): 9857\n```\nThen it can be easier to understand what network calls the API Client is making, and where any issues may exist. If you\nwish to dig even deeper, you may consider modifying the `_send_request()` method of the API Client to log additional \ndata about the state of your environment, or anything else that would help diagnose issues. You will notice a log file\ngenerated in the format `cm_api_client_debug_2023_01_09_11_01_02.txt`. This log file might be helpful for your own use\nor to give more context if you are working with Coin Metrics customer success. \n\n### SSL Certs verification\n\nSometimes your organization network have special rules on SSL certs verification and in this case you might face the\nfollowing error when running the script:\n```text\nSSLError: HTTPSConnectionPool(host=\'api.coinmetrics.io\', port=443): Max retries exceeded with url: <some_url_path> (Caused by SSLError(SSLCertVerificationError(1, \'[SSL: CERTIFICATE_VERIFY_FAILED] certificate verify failed: self signed certificate in certificate chain (_ssl.c:1123)\')))\n```\n\nIn this case, you can pass an option during client initialization to disable ssl verification for requests like this:\n\n```python\n\nclient = CoinMetricsClient(verify_ssl_certs=False)\n```\n\nWe don\'t recommend setting it to False by default and you should make sure you understand the security risks of disabling SSL certs verification.\n\nAdditionally, you may choose to specify the path to the SSL certificates on your machine. This may cause errors where \nPython is unable to locate the certificates on your machine, particularly when using Python virtual environments. \n\n```python\nfrom coinmetrics.api_client import CoinMetricsClient\nSSL_CERT_LOCATION = \'/Users/<USER_NAME>/Library/Python/3.8/lib/python/site-packages/certifi/cacert.pem\'\nclient = CoinMetricsClient(verify_ssl_certs=SSL_CERT_LOCATION)\n```\n\nA quick way to find the certs on your machine is:  \n`python3 -c "import requests; print(requests.certs.where())"`  \nAnd note that this will change based on whether or not you are using a [Python virtual environment or not](https://realpython.com/python-virtual-environments-a-primer/)\n\n### Installing and running coinmetrics package and other python packages behind a secure python network\nRelated to SSL Certs verification, you may have trouble installing and updating PyPi packages to your local environment.\nSo you may need to choose the best solution for your company and environment - either using package managers or\ninstalling offline.\n\n#### Installing using package managers\nFull instructions for setting up your environment to use conda, pip, yarn, npm, etc. can be [found here](https://medium.com/@iffi33/dealing-with-ssl-authentication-on-a-secure-corporate-network-pip-conda-git-npm-yarn-bower-73e5b93fd4b2).\nAdditionally, a workaround to disable SSL verification when installing a trusted Python package is this:  \n```commandline\npip install --trusted-host pypi.python.org <packagename>\n```  \nAlthough it is important to make sure you understand the risks associated with disabling SSL verification and ensure \ncompliance with company policies.\n\n\n\n#### Installing Python packages locally/ offline\nIt may be easier to download and install the package locally. Steps:  \n1. Download the files for the [Coin Metrics API Client from PyPi](https://pypi.org/project/coinmetrics-api-client/#files)\n2. [Install it locally](https://packaging.python.org/en/latest/tutorials/installing-packages/#installing-from-local-archives)\n\n### Requests Proxy\nSometimes your organization has special rules on making requests to third parties and you have to use proxies in order to comply with the rules.\n\nFor proxies that don\'t require auth you can specify them similar to this example:\n```python\n\nclient = CoinMetricsClient(proxy_url=f\'http://<hostname>:<port>\')\n```\n\nFor proxies that require auth, you should be able to specify username and password similar to this example:\n```python\n\nclient = CoinMetricsClient(proxy_url=f\'http://<username>:<password>@<hostname>:<port>\')\n```\n\n## Extended documentation\nFor more information about the available methods in the client please reference [API Client Spec](https://coinmetrics.github.io/api-client-python/site/api_client.html)\n\n',
    'author': 'Coin Metrics',
    'author_email': 'info@coinmetrics.io',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://coinmetrics.github.io/api-client-python/site/index.html',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'entry_points': entry_points,
    'python_requires': '>=3.7.1,<4.0.0',
}


setup(**setup_kwargs)



# type: ignore
import pytest
import orjson
import os
from coinmetrics.api_client import CoinMetricsClient, CmStream

client = CoinMetricsClient(str(os.environ.get("CM_API_KEY")))
cm_api_key_set = os.environ.get("CM_API_KEY") is not None
REASON_TO_SKIP = "Need to set CM_API_KEY as an env var in order to run this test"

print("CM_API_KEY is set - tests will run") if cm_api_key_set else print(
    "CM_API_KEY not set, tests will not run"
)


def on_message_index_levels_test(stream: CmStream, message: str) -> None:
    """
    Tests that data with the expected keys can be loaded by orjson
    """
    data = orjson.loads(message)
    expected_cols_index_levels = ["index", "time", "level", "cm_sequence_id"]
    for col in expected_cols_index_levels:
        assert col in data
    stream.close()


def on_message_market_trades_test(stream: CmStream, message: str) -> None:
    """
    Tests that data with the expected keys can be loaded by orjson
    """
    data = orjson.loads(message)
    expected_cols_index_levels = [
        "market",
        "time",
        "coin_metrics_id",
        "amount",
        "price",
        "collect_time",
        "side",
        "cm_sequence_id",
    ]
    for col in expected_cols_index_levels:
        assert col in data
    stream.close()


def on_message_market_orderbooks_test(stream: CmStream, message: str) -> None:
    """
    Tests that data with the expected keys can be loaded by orjson
    """
    data = orjson.loads(message)
    expected_cols_index_levels = [
        "market",
        "time",
        "coin_metrics_id",
        "asks",
        "bids",
        "type",
        "collect_time",
        "cm_sequence_id",
    ]
    for col in expected_cols_index_levels:
        assert col in data
    stream.close()


def on_message_market_candles_test(stream: CmStream, message: str) -> None:
    """
    Tests that data with the expected keys can be loaded by orjson
    """
    data = orjson.loads(message)
    expected_cols_index_levels = [
        "market",
        "time",
        "price_open",
        "price_close",
        "price_high",
        "price_low",
        "vwap",
        "volume",
        "candle_usd_volume",
        "candle_trades_count",
        "cm_sequence_id",
    ]
    for col in expected_cols_index_levels:
        assert col in data
    stream.close()


def on_message_market_quotes_test(stream: CmStream, message: str) -> None:
    """
    Tests that data with the expected keys can be loaded by orjson
    """
    data = orjson.loads(message)
    expected_cols_index_levels = [
        "market",
        "time",
        "coin_metrics_id",
        "ask_price",
        "ask_size",
        "bid_price",
        "bid_size",
        "cm_sequence_id",
    ]
    for col in expected_cols_index_levels:
        assert col in data
    stream.close()


def on_message_asset_metrics_rr_test(stream: CmStream, message: str) -> None:
    """
    Tests that data with the expected keys can be loaded by orjson
    """
    data = orjson.loads(message)
    expected_cols_index_levels = ["time", "asset", "ReferenceRateUSD", "cm_sequence_id"]
    for col in expected_cols_index_levels:
        assert col in data
    stream.close()


@pytest.mark.skipif(not cm_api_key_set, reason=REASON_TO_SKIP)
def test_index_levels_stream() -> None:
    indexes = ["CMBIBTC", "CMBIETH"]
    stream = client.get_stream_index_levels(indexes=indexes)
    stream.run(on_message=on_message_index_levels_test)


@pytest.mark.skipif(not cm_api_key_set, reason=REASON_TO_SKIP)
def test_market_trades_stream() -> None:
    markets = ["coinbase-btc-usd-spot"]
    stream = client.get_stream_market_trades(markets=markets)
    stream.run(on_message=on_message_market_trades_test)


@pytest.mark.skipif(not cm_api_key_set, reason=REASON_TO_SKIP)
def test_market_orderbooks_stream() -> None:
    markets = ["binance-btc-usdt-spot"]
    stream = client.get_stream_market_orderbooks(markets=markets)
    stream.run(on_message=on_message_market_orderbooks_test)


@pytest.mark.skipif(not cm_api_key_set, reason=REASON_TO_SKIP)
def test_market_quotes_stream() -> None:
    markets = ["*"]
    stream = client.get_stream_market_quotes(markets=markets)
    stream.run(on_message_market_quotes_test)


@pytest.mark.skipif(not cm_api_key_set, reason=REASON_TO_SKIP)
def test_asset_metrics() -> None:
    assets = ["btc"]
    metrics = ["ReferenceRateUSD"]
    stream = client.get_stream_asset_metrics(
        assets=assets, metrics=metrics, frequency="1s"
    )
    stream.run(on_message_asset_metrics_rr_test)


@pytest.mark.skipif(not cm_api_key_set, reason=REASON_TO_SKIP)
def test_market_candles_stream() -> None:
    stream = client.get_stream_market_candles(
        markets=["coinbase-btc-usd-spot"], frequency="1m"
    )
    stream.run(on_message_market_candles_test)
